%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% The "calcio" multiagent domain  %%%%
%%%% by DFP - 2011                   %%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% WHITE TEAM
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

team(black). team(white).
num(1). num(2). num(3).
color(black,1).    color(white,0).

agent(player(white,X)) :- num(X).
agent(player(black,X)) :- num(X).

%%% Admissible directions and corresponding offsets

delta(nw,-1,1).  delta(n,0,1).  delta(ne,1,1).
delta( w,-1,0).                 delta(e,1,0).
delta(sw,-1,-1). delta(s,0,-1). delta(se,1,-1).

direction(D) :- delta(D,_,_).

%%% Admissible kick strength

power(X) :- interval(X,2,3).


xmax(7).  ymax(4).
xfullpos(X) :- xmax(M), interval(X,0,M).
xpos(X) :- xmax(M), M1 is M-1, interval(X,1,M1).
ypos(X) :- ymax(M), interval(X,0,M).

%positions of the two "goals"
door(white,0,D) :- ymax(M), D is M//2.
door(black,Mx,D) :- ymax(My), D is My//2, xmax(Mx).

%%%%%%  FLUENTS
fluent(x(ball) , 0,M) :- xmax(M).
fluent(y(ball) , 0,M) :- ymax(M).
fluent(x(A), 1,M1) :- agent(A), xmax(M), M1 is M-1.
fluent(y(A), 0,M) :- agent(A), ymax(M).
fluent(hasball(A),0,1) :- agent(A).
fluent(cell(X,Y),-1,1) :- xfullpos(X), ypos(Y).
     % 0,1 = occupied by black,white,  -1=free
%fluent(win(T),    0,1) :- team(T).


%%%%%
action(act([player(white,A)],move(D))) :-
    agent(player(white,A)),direction(D).

action(act([player(white,A)],kick(D,F))) :-
    agent(player(white,A)),direction(D),power(F).

action(act([player(white,A)],takeball)) :-
    agent(player(white,A)).

%%% Executability conditions
%%% The player cannot exit, cannot enter in door
executable  act([A],move(D))  if  [1 leq x(A)+DX, x(A)+DX leq M1, 0 leq y(A)+DY, y(A)+DY leq My | Cs]  :-
    action(act([A],move(D))), delta(D,DX,DY),
    ymax(My), xmax(Mx), M1 is Mx-1,
    %notOccupied(A,D,Cs).
    notOtherPlayers(A,D,Cs).

executable  act([A],takeball)  if  [ pair(x(A),y(A)) eq pair(x(ball),y(ball)) , hasball(A) eq 0]  :-
    action(act([A],takeball)).

%notOccupied(A,D, [cell(x(A)+DX,y(A)+DY) eq -1] ) :-
%       delta(D,DX,DY).

notOtherPlayers(A,D,Cs) :-
       delta(D,DX,DY),
       findall((pair(x(A)+DX,y(A)+DY) neq pair(x(B),y(B))), (agent(B), neq(A,B)) , Cs).


%%% The ball cannot exit

executable  act([A],kick(D,F))  if  [hasball(A) eq 1,
                        1 leq x(A)+DX*F, x(A)+DX*F leq M1, 0 leq y(A)+DY*F, y(A)+DY*F leq My]  :-
    action(act([A],kick(D,F))),delta(D,DX,DY),
    ymax(My), xmax(Mx), M1 is Mx-1.


%%% But can enter the proper door

executable  act([player(white,N)],kick(D,F))   if  [hasball(player(white,N)) eq 1,
           pair(x(player(white,N))+DX*F, y(player(white,N))+DY*F) eq pair(X,Y)]   :-
    action(act([player(white,N)],kick(D,F))),delta(D,DX,DY),
    door(white,X,Y).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Effects of move:
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

[actocc([A],move(D))]  causes (pair(x(A),y(A)) eq pair(x(A)^(-1)+DX, y(A)^(-1)+DY)) :-
    action(act([A],move(D))), delta(D,DX,DY).
[actocc([A],move(D)), pair(x(A),y(A)) eq pair(X,Y) ]  causes (cell(X,Y) eq Col) :-
    action(act([A],move(D))), A=player(Team,_), color(Team,Col),
    fluent(cell(X,Y)).
[actocc([A],move(D)), pair(x(A),y(A)) eq pair(X,Y) ]  causes (cell(X,Y)^(-1) eq -1) :-
    action(act([A],move(D))),
    fluent(cell(X,Y)).
[actocc([A],move(D)), hasball(A) eq 1]   causes (pair(x(ball),y(ball)) eq pair(x(A)^(-1)+DX, y(A)^(-1)+DY)) :-
    action(act([A],move(D))), delta(D,DX,DY).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Effects of takeball:
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

[actocc([A],takeball)]  causes (hasball(A) eq 1) :-
    action(act([A],takeball)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Effects of Kick:
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%% Once agent A has kicked, he has the ball nomore

[actocc([A],kick(D,F))]  causes  (hasball(A) eq 0) :-
    action(act([A],kick(D,F))).

%%% Handling of passage. The ball reaches the desired cell.
[actocc([Sender],kick(D,F))]  causes  
(pair(x(ball),y(ball)) eq pair(x(Sender)^(-1)+F*DX,y(Sender)^(-1)+F*DY)) :-
    action(act([Sender],kick(D,F))), delta(D,DX,DY).


%subsumed by (1)
%always(pair(x(A),y(A)) neq pair(x(B),y(B))) :-
%       agent(A), agent(B), neq(A,B).

%%% (1)
%%% roles for players
% 1 is defender/goalkeeper, 2 is midfield, 3 is attack
always( x(player(white,1)) gt x(player(white,2)) ).
always( x(player(white,2)) gt x(player(white,3)) ).
%goalkeeper must stay close to his door (the goal of the opponent)
always( (X - x(player(white,1)))  + abs(y(player(white,1))-Y) leq 2 ) :- door(black,X,Y).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Initial state
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%        players   1         2        3
blackpositions([pair(1,2),pair(2,4),pair(5,0)]).
whitepositions([pair(6,2),pair(5,4),pair(4,0)]).

initially(pair(x(player(black,I)),y(player(black,I))) eq Pos) :- blackpositions(Black), nth1(I,Black,Pos).
initially(pair(x(player(white,I)),y(player(white,I))) eq Pos) :- whitepositions(White), nth1(I,White,Pos).
%%%
initially(pair(x(ball),y(ball)) eq pair(6,2)).

%initially(win(T) eq 0) :- team(T).
initially(hasball(player(black,A)) eq 0) :- num(A).
initially(hasball(player(white,1)) eq 1).
initially(hasball(player(white,2)) eq 0).
initially(hasball(player(white,3)) eq 0).

initially(cell(X,Y) eq 1) :- blackpositions(Black), member(pair(X,Y),Black).
initially(cell(X,Y) eq 0) :- whitepositions(White), member(pair(X,Y),White).
initially(cell(X,Y) eq -1) :-
                              blackpositions(Black),
                              whitepositions(White),
                              xfullpos(X),ypos(Y),
                              nonmember(pair(X,Y),Black),
                              nonmember(pair(X,Y),White).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% GOAL
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%Goal for this team: to score
goal(pair(x(ball),y(ball)) eq pair(X,Y)) :- door(white,X,Y).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% AD HOC PRINTING OF THE RESULT
%%% %%% %%% %%% %%% %%% %%% %%%

myprint(S) :- 
     nl,
     soccer_print(0,S).

soccer_print(_,[]).
soccer_print(N,[S|Rest]) :-
     write('Time '),write(N),write(': '),nl,
     write_campo(S),
     N1 is N + 1,
     soccer_print(N1,Rest).

write_campo(S) :-
    xmax(MX),X is MX+1,
    ymax(MY),Y is MY+2,
    build_field(0,Y,X,FIELD),
    findall(P,agent(P),L),
    assign_players(L,S,FIELD),
    assign_ball(FIELD,S),
    %
    borders(FIELD),
    reverse(FIELD,REV),
    stampamatrice(REV).

build_field(Y,MY,_,[]) :- Y>MY,!.
build_field(Y,MY,MX,[ROW|FIELD]) :-
     length(ROW,MX),
     Y1 is Y+1,
     build_field(Y1,MY,MX,FIELD).


assign_players([],_,_).
assign_players([player(T,N)|R],S,FIELD) :-
     member(fluent(x(player(T,N)),WX),S),
     member(fluent(y(player(T,N)),WY),S),
     (member(fluent(hasball(player(T,N)),1),S),!,
       insertoken(T,[WX,WY,1],FIELD);
     insertoken(T,[WX,WY,0],FIELD)),
     assign_players(R,S,FIELD).

assign_ball(FIELD,S) :-
     member(fluent(x(ball),WX),S),
     member(fluent(y(ball),WY),S),
     insertoken(ball,[WX,WY,0],FIELD).

insertoken(TEAM,[X,Y,F],MAT) :-
     (
     (var(X);var(Y)),!,write(player(TEAM,undefined)),nl;
     Y1 is Y+1, nth0(Y1,MAT,L),
     nth0(X,L,VAR),
     (var(VAR),!,
       ((TEAM=white,F=0) -> VAR='O';
        (TEAM=white,F=1) -> VAR='Q';
        (TEAM=black,F=0) -> VAR='Y';
        (TEAM=black,F=1) -> VAR='X';
        (TEAM=net,F=0) -> VAR='|';
        (TEAM=net,F=1) -> VAR='o';
        (TEAM=ball)      -> VAR='o');
      true),
     !
     ).

borders(FIELD) :-
   length(FIELD,N),
   nth1(1,FIELD,U),
   nth1(N,FIELD,O),
   append([U,O],Bord),
   stelline(Bord),
   %%% Here to be refined:
   ymax(M), Door is M//2+1,
   left(FIELD,L),
   stelline(L,0,Door),
   right(FIELD,R),
   stelline(R,0,Door).

left([],[]).
left([[A|_]|R],[A|S]) :- left(R,S).

right([],[]).
right([A|R],[B|S]) :-
   reverse(A,[B|_]), right(R,S).

stelline([]).
stelline([A|R]) :-
   (var(A) -> A='*'; true),
   stelline(R).

stelline([],_,_).
stelline([_|R],K,K) :-
   !,
   stelline(R).
stelline([A|R],K,H) :-
   (var(A) -> A='*'; true),
   K1 is K+1,
   stelline(R,K1,H).


%%%% AUX %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

stampamatrice([]).
stampamatrice([A|R]) :-
      stampariga(A),nl, stampamatrice(R).
stampariga([]).
stampariga([A|R]) :-
     (var(A),!, write(' '); write(A)), stampariga(R).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

