%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% The "volley" multiagent domain  %%%%
%%%% by DFP - 2011                   %%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% BLACK TEAM
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%
%%% ************* %%%
%%% *X    |    O* %%%
%%% *     |     * %%%  %%%
%%% *     |     * %%%  %%% Y axis
%%% *     |     * %%%  %%%
%%% *Y    |    O* %%%  %%% (0-out)1..5(6-out)
%%% ************* %%%
%%%%%%%%%%%%%%%%%%%%%
%%% X axis (0-out)1..5(6)7..11(12-out) %%%

%%% Legenda:
%%% black -> X, white -> O
%%% If they have ball: Y and Q.

%%% Field size

linex(12). %%% Must be even
net(6).    %%% net is linex/2
liney(6).  %%% Can be even or odd

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

team(black). team(white).
num(1).      num(2).
opposite(black,white). opposite(white,black).
power(X) :- interval(X,1,5). %linex(BIG),M is BIG-2,interval(X,2,M).

%%% Admissible directions and offsets:

delta(nw,-1,1).  delta(n,0,1).  delta(ne,1,1).
delta(w,-1,0).   delta(j,0,0).  delta(e,1,0).
delta(sw,-1,-1). delta(s,0,-1). delta(se,1,-1).

direction(D) :- delta(D,_,_).

%%% Agents

agent(player(T,X)) :- team(T),num(X).
%agent(referee).
player(A) :- agent(A). %, neq(A,referee).
blackplayer(A) :- player(A), A=player(black,_).

%%%%%%  FLUENTS

fluent(x(player(white,X)),B,E) :-
      num(X), net(NET),B is NET+1,
      linex(MX),E is MX-1.
fluent(x(player(black,X)),1,E) :-
      num(X), net(NET),E is NET-1.
fluent(y(A),1,M) :-
      player(A), liney(MY),M is MY-1.

%%%%%%%

fluent(x(ball),1,MX1) :- linex(MX), MX1 is MX-1. %linea = fuori
fluent(y(ball),1,MY1) :- liney(MY), MY1 is MY-1.
%fluent(x(ball),0,MX) :- linex(MX).
%fluent(y(ball),0,MY) :- liney(MY).

%%%%%%%

fluent(hasball(A),0,1) :- player(A).
fluent(point(T),0,1) :- team(T).
fluent(defense_time,0,1). %%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%% - move (like a chess king). If it has not the ball, if
%%%   it is defense time
%%% - throw (the ball goes like a queen, for F steps)

action(act([A],move(D))) :-
     blackplayer(A),direction(D).

action(act([A],throw(D,F))) :-
     blackplayer(A),direction(D), neq(D,j), power(F).

action(act([player(black,1)],fischio)).


executable act([player(black,1)],fischio) if [S eq 0] :-
      findall(hasball(P),player(P),LP),
      build_sum(LP,S).

build_sum([], defense_time).
build_sum([A|R],B) :- build_sum(R,C), B =.. [+,A,C].


%%% Executability conditions

%%% Observe the defense_time constraint

%%% The player cannot exit

executable  act([A],move(D))  if [hasball(A) eq 0, defense_time gt 0,
                         0 lt  x(A)+DX, x(A)+DX lt MX,
                         0 lt  y(A)+DY, y(A)+DY leq MY] :-
      action(act([A],move(D))), delta(D,DX,DY),
      net(MX), liney(MY).

%%% The ball cannot exit (but can enter in door).
%%% Nota: si puo' tirare sulla riga (e si fa punto!)
%%% e anche sulla rete (e si perde). Ma non insegniamo
%%% a tirare sulla riga, vediamo se lo capisce da solo.

executable  act([A],throw(D,F))  if  [hasball(A) gt 0,defense_time eq 0,
                         0 lt x(A)+DX*F, x(A)+DX*F lt MX,
                         0 lt y(A)+DY*F, y(A)+DY*F lt MY]  :-
     action(act([A],throw(D,F))),delta(D,DX,DY),
     linex(MX), liney(MY).

%%% In pratica:
%%% hasball(player(white,1)) + ... + hasball(player(white,2)) +
%%% defense_time eq 0

[actocc([player(black,1)],fischio), 0 leq x(ball), x(ball) leq N]
causes
point(white) eq 1 :-
        net(NET), N is NET-1.

[actocc([player(black,1)],fischio), N leq x(ball), x(ball) leq MX]
causes
point(black) eq 1 :-
        net(NET), N is NET+1,
        linex(MX).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Effects of move:
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

[actocc([A],move(D))]
causes
pair(x(A),y(A)) eq pair(x(A)^(-1)+DX,y(A)^(-1)+DY) :-
      action(act([A],move(D))), delta(D,DX,DY).

%%% Defense time is decreased

[actocc([A],move(D))]
causes
defense_time eq defense_time^(-1)-1 :-
     action(act([A],move(D))).

%%% Qui cattura la palla spostandosi in tempo prima che cada

%[actocc([A],move(D)), pair(x(ball)^(-1),y(ball)^(-1)) eq pair(x(A)^(-1)+DX,y(A)^(-1)+DY)]
[actocc([A],move(D)), pair(x(ball),y(ball)) eq pair(x(A)+DX,y(A)+DY)]
causes
hasball(A) eq 1 :-
      action(act([A],move(D))), delta(D,DX,DY).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Effect of throw
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%% Se tira non ha piu' la palla

[actocc([A],throw(D,F))]
causes
hasball(A) eq 0
:-
     action(act([A],throw(D,F))).

%%% Quando tira parte defense_time in cui ci si puo' spostare

[actocc([A],throw(D,F))]
causes
defense_time eq 1
:-
     action(act([A],throw(D,F))).

%%% La palla si sposta

[actocc([A],throw(D,F))]
causes
pair(x(ball),y(ball)) eq  pair(x(A)^(-1) + F*DX,y(A)^(-1) + F*DY)
     :-
     action(act([A],throw(D,F))), delta(D,DX,DY).

% Nota: quando uno tira, defense_time e' a zero dunque nessun si sposta.
% (a parte ... da scoprire)
% Se uno e' nel posto giusto piglia la palla

%[actocc([A],throw(D,F)),
%      pair(x(B)^(-1),y(B)^(-1)) eq
%      pair(x(A)^(-1)+F*DX, y(A)^(-1)+F*DY)]
%causes hasball(B) eq 1 :- action(act([A],throw(D,F))), player(B), neq(A,B),delta(D,DX,DY).


%%% Se butta palla in rete gli altri fan punto

     [actocc([A],throw(D,F)), x(A)+F*DX eq Net]
causes
point(white) eq 1
     :-
     action(act([A],throw(D,F))), delta(D,DX,_),
     net(Net).

%%%%%%%%%%%%%%%%%%%%%%%%%%%

always(pair(x(A),y(A)) neq pair(x(B),y(B))) :-
     A=player(black,N),B=player(black,M), num(N), num(M), N<M.

%%% Initial situation

initially(x(player(black,1)) eq 1).
initially(y(player(black,1)) eq 1).
initially(x(player(black,2)) eq 3).
initially(y(player(black,2)) eq 4).

%

initially(x(player(white,1)) eq 9).
initially(y(player(white,1)) eq 2).
initially(x(player(white,2)) eq 11).
initially(y(player(white,2)) eq 4).

%

initially(x(ball) eq 1). initially(y(ball) eq 1).
initially(hasball(player(black,1)) eq 1).
initially(hasball(AG) eq 0) :-
         player(AG),
         \+ initially(hasball(AG) eq 1).

initially(point(T) eq 0) :- team(T).

initially(defense_time eq 0).

%%%%

goal(point(black) eq 1).
goal(point(white) eq 0).

%%% %%% %%% %%% %%% %%% %%% %%%
%%% PRINTING AD HOC
%%% %%% %%% %%% %%% %%% %%% %%%

myprint(S) :- volley_print(0,S).

volley_print(_,[]).
volley_print(N,[S|Rest]) :-
     write('Time '),write(N),write(': '),nl,
     write_campo(S),
     N1 is N + 1,
     volley_print(N1,Rest).

build_field(Y,MY,_,[]) :- Y>MY,!.
build_field(Y,MY,MX,[ROW|FIELD]) :-
     length(ROW,MX),
     Y1 is Y+1,
     build_field(Y1,MY,MX,FIELD).

write_campo(S) :-
    linex(MX),X is MX+1,
    liney(MY),
    net(NET),
    build_field(0,MY,X,FIELD),
    findall(P,player(P),L),
    assign_players(L,S,FIELD),
    assign_ball(FIELD,S),
    paint_net(MY,FIELD,NET,S),
    borders(FIELD),
    reverse(FIELD,REV),
    stampamatrice(REV).

assign_players([],_,_).
assign_players([player(T,N)|R],S,FIELD) :-
     member(fluent(x(player(T,N)),WX),S),
     member(fluent(y(player(T,N)),WY),S),
     (member(fluent(hasball(player(T,N)),1),S),!,
       insertoken(T,[WX,WY,1],FIELD);
     insertoken(T,[WX,WY,0],FIELD)),
     assign_players(R,S,FIELD).

assign_ball(FIELD,S) :-
     member(fluent(x(ball),WX),S),
     member(fluent(y(ball),WY),S),
     insertoken(ball,[WX,WY,0],FIELD).

paint_net(-1,_,_,_):-!.
paint_net(Y,FIELD,NET,S) :-
     (member(fluent(x(ball),NET),S),
      member(fluent(y(ball),Y),S),!,
      insertoken(net,[NET,Y,1],FIELD);
      insertoken(net,[NET,Y,0],FIELD)),
     Y1 is Y-1,
     paint_net(Y1,FIELD,NET,S).


insertoken(TEAM,[X,Y,F],MAT) :-
     (
     (var(X);var(Y)),!,write(player(TEAM,undefined)),nl;
     nth0(Y,MAT,L),
     nth0(X,L,VAR),
     (var(VAR),!,
       ((TEAM=white,F=0) -> VAR='O';
        (TEAM=white,F=1) -> VAR='Q';
        (TEAM=black,F=0) -> VAR='Y';
        (TEAM=black,F=1) -> VAR='X';
        (TEAM=net,F=0) -> VAR='|';
        (TEAM=net,F=1) -> VAR='o';
        (TEAM=ball)      -> VAR='o');
      true),
     !
     ).

borders(FIELD) :-
   length(FIELD,N),
   nth1(1,FIELD,U),
   nth1(N,FIELD,O),
   left(FIELD,L),
   right(FIELD,R),
   append([U,O,L,R],Bord),
   stelline(Bord).

left([],[]).
left([[A|_]|R],[A|S]) :- left(R,S).

right([],[]).
right([A|R],[B|S]) :-
   reverse(A,[B|_]), right(R,S).

stelline([]).
stelline([A|R]) :-
   (var(A) -> A='*'; true),
   stelline(R).


%%%% AUX %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

stampamatrice([]).
stampamatrice([A|R]) :-
      stampariga(A),nl, stampamatrice(R).
stampariga([]).
stampariga([A|R]) :-
     (var(A),!, write(' '); write(A)), stampariga(R).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
