:- module(arbitration_opt, [solveByArbitration/4]).
% DA DEBUG-ARE
%:- module(arbitration_opt, [test/4,solveByArbitration/4]).

%
% grezza possibile implementazione dell'ARBITRATION:
% dato un insieme di azioni in conflitto,
% cerca con CLP una soluzione che minimizzi il numero di azioni
% da bloccare.

:- use_module(library(lists)).
:- use_module(library(clpfd)).


solveByArbitration(PTACsList, TckAOptList, Goods, Bads) :-
	setof(T-(A-Cs), (member(_-T-(A-Cs), PTACsList),
			 (member(T-A-'RETRY-OP',TckAOptList) ; member(T-A-'ARBITRATE',TckAOptList))
                        ),     %                                  ^ CONTROLLARE ^
		NotForegone_TACss),!,
	(setof(P-T-(A-_Cs), (member(P-T-(A-_Cs),PTACsList),
			      nonmember(T-(A-_), NotForegone_TACss)), Bads0)
	; Bads0=[]),!,
	findConf(NotForegone_TACss, As, VarAs, _Fs, _VarFs),!,
	splitGoodsBads(As, VarAs, PTACsList, Goods, Bads1),
	append(Bads0, Bads1, Bads).
solveByArbitration(PTACsList, _TckAOptList, [], PTACsList).


splitGoodsBads([], [], _PTCsList, [], []).
splitGoodsBads([T-A|TAs], [VarTA|VarTAs], PTACsList, [P-T-(A-Cs)|Goods], Bads) :-
	VarT == 1, !,
	memberchk(P-T-(A-Cs), PTACsList),
	splitGoodsBads(TAs, VarTAs, PTACsList, Goods, Bads).
splitGoodsBads([T-A|TAs], [_VarTA|VarTAs], PTACsList, Goods, [P-T-(A-Cs)|Bads]) :-
	memberchk(P-T-(A-Cs), PTACsList),
	splitGoodsBads(TAs, VarTAs, PTACsList, Goods, Bads).


% trova una soluzione inibendo un numero minimale di agenti
% per ogni conflitto sceglie uno solo degli agent-act in conflitto
%findConf(+TACsList, -As, -VarAs, -Fs, -VarFs) 
findConf(TACsList, As, VarAs, Fs, VarFs) :-
	effects(TACsList, Quads, AgAct_s, Fs, MinVal, MaxVal),
	length(AgAct_s, NumAs),
	length(VarAs, NumAs),
	%write(VarAs),
	domain(VarAs, 0 , 1),
	length(Fs, NumFs),
	length(VarFs, NumFs),
	%write(VarFs),
	domain(VarFs, MinVal, MaxVal),
	constrainer(Quads, AgAct_s, VarAs, Fs, VarFs),
	sum(VarAs, #=, NumActing),
	labeling([maximize(NumActing)],VarAs).



%effects(+TACsList, -Quads, -OrdAAs, -OrdFs, -MinVal, -MaxVal)
effects(TACsList, Quads, OrdAAs, OrdFs, MinVal, MaxVal) :-
	findall((T-Act)-(F-V),(member(T-(Act-Cs), TACsList), member([F,V],Cs)),Quads),
	keys_and_values(Quads, AgentsActs, FVs),
	keys_and_values(FVs, Fluents, Vals),
	sort(AgentsActs, OrdAAs),
	sort(Fluents, OrdFs),
	max_member(MaxVal,Vals),
	min_member(MinVal,Vals).

constrainer([], _, _, _, _).
constrainer([(T-A)-(F-V)|Quads], AgAct_s, VarAs, Fs, VarFs) :-
	nth0(PosA, AgAct_s, (T-A)), nth0(PosA, VarAs, VarA),
	nth0(PosF, Fs, F), nth0(PosF, VarFs, VarF),
	(VarA #=> (VarF#=V)),
	constrainer(Quads, As, VarAs, Fs, VarFs).








%TEST CODE:

%getActionEffects(tik1,[[a1,100],[c1,1500]]).
%getActionEffects(tik2,[[b2,200]]).
%getActionEffects(tik3,[[a3,-100],[b3,200]]).
getActionEffects(tik4,[[conf,300],[b4,200]]).
getActionEffects(tik5,[[conf,200],[b5,200]]). %conflict with 4
%getActionEffects(tik5,[[conf,300],[b5,200]]).  %no conflict

getActionEffects(tik6,[[a6,106],[b6,206]]). %
getActionEffects(tik7,[[a6,107],[b8,207]]). % trio in conflitto
getActionEffects(tik8,[[a8,108],[b8,208]]). %

%getActionEffects(tik9,[[a9,109],[b9,209]]).
%getActionEffects(tik10,[[a10,110],[b10,210]]).
%getActionEffects(tik11,[[a11,111],[b11,211]]).
%getActionEffects(tik12,[[a12,112],[b12,212]]).

test(PTCsList, TckOptList, Goods, Bads) :-
	findall(1-T-Cs,getActionEffects(T,Cs),PTCsList),
	findall(T-'RETRY-OP', member(_-T-_, PTCsList), TckOptList),
	solveByArbitration(PTCsList, TckOptList, Goods, Bads).


