:- module(client, [startConnection/2,closeConnection/1,setInitialState/5,performOneStep/4,terminalStep/1,noOpStep/1]).

:- use_module('settings.pl').
:- use_module(library(lists)).
:- use_module(m_lindaclient).

% different ways to resolve conflicts (see settings.pl)
:- conflictResolutionStrategy(_N, _, ClientModuleFile),
   use_module(ClientModuleFile).




%%%%%%%%%%%%%%
%%TODO:
%%%%%%%%%%%%%%%
%
% La attuale 
% processResponse/6
% stampa un messaggio che potrebbe essere inibito. Da valutare se oltre a
% restituire il responso (ultimo argomento) debba fare qualcos'altro....



terminalStep(AgentTck) :-
	m_rd(startStep(StepNum)),
	( (StepNum == interactionEnded) 
		-> ( format("AgentTck ~w : got interactionEnded ~n", [AgentTck])
			% some housekeeping ?...
		)
		;  ( format("AgentTck ~w : got StepNum = ~w ~n", [AgentTck,StepNum]),
		     format("AgentTck ~w : sending termination... ", [AgentTck]),
		     m_out(proposedAction(StepNum,AgentTck,'NO-MORE')),
		     format("...sent:~w.~n", [proposedAction(StepNum,AgentTck,'NO-MORE')])
		)
	).



noOpStep(AgentTck) :-
	m_rd(startStep(StepNum)),
	( (StepNum == interactionEnded) 
		-> ( format("AgentTck ~w : got interactionEnded ~n", [AgentTck])
			% some housekeeping ?...
		)
		;  ( format("AgentTck ~w : got StepNum = ~w ~n", [AgentTck,StepNum]),
		     format("AgentTck ~w : sending NO-OP... ", [AgentTck]),
		     m_out(proposedAction(StepNum,AgentTck,'NO-OP')),
		     format("...sent:~w.~n", [proposedAction(StepNum,AgentTck,'NO-OP')])
		)
	).



% POSSIBILI AZIONI :
% % 'NO-OP'    =  l'agente in questo turno attende
% % 'NO-MORE'  =  l'agente ha terminato
% ActionNameMulti e' una lista di oggetti della forma ActName-AffectefFluents, dove ActName e' l'azione
% e AffectefFluents e' una lista dei [V,F] influenzati da ActName nella transizione
performOneStep(AgentTck,ActionNameMulti,_APairs,Responses) :-
	%write('STEP ME client:performOneStep, (type an atom):'),read(_),
	m_rd(startStep(StepNum)),
	( (StepNum == interactionEnded) 
		-> ( Responses=[interactionEnded], format("AgentTck ~w : got interactionEnded ~n", [AgentTck])
			% some housekeeping ?...
		)
		;  ( format("AgentTck ~w : got StepNum = ~w ~n", [AgentTck,StepNum]),
		     length(ActionNameMulti,Num),
		     format("AgentTck ~w : sending ~w action(s)... ", [AgentTck,Num]),
     %write('STEP ME client: before sendMicroActions, (type an atom):'),read(_),nl,
		     sendMicroActions(ActionNameMulti, AgentTck,StepNum),
		     format("...sent.  Waiting for response(s)...~n", []),
		     getMicroActionsResponses(ActionNameMulti, 1, Num, AgentTck, StepNum, Respns_s),
		     processResponses(Respns_s,StepNum,AgentTck,ActionNameMulti,[],Responses)  %***
		)
	).

getMicroActionsResponses([],N1,N,_AgentTck,_StepNum,[]) :- N1 =:= N+1,!.
getMicroActionsResponses([],N1,N,AgentTck,StepNum,[]) :- N1 =\= N+1,!, 
	format(" ~n***** UNEXPECTED ERROR: Agent ~w: wrong number of actions in step~w.~n~n", [AgentTck,StepNum]).
getMicroActionsResponses([ActName-_MicroDelta|Rest],N,M,AgentTck,StepNum,[Respns|Respns_s]) :-  
	m_in(actionResponse(StepNum,AgentTck,Respns)),
	format("              received action response:~w --- ~w~n", [ActName,Respns]),
	N1 is N+1,
	getMicroActionsResponses(Rest,N1,M,AgentTck,StepNum,Respns_s).


%sendMicroActions([],N1,N,AgentTck,StepNum) :- N1 =:= N+1,!,
%	m_out(proposedActionCount(StepNum,AgentTck,N)).
%sendMicroActions([],N1,N,AgentTck,StepNum) :- N1 =\= N+1,!, 
%	format(" ~n***** UNEXPECTED ERROR: Agent ~w: wrong number of actions in step~w.~n~n", [AgentTck,StepNum]).
sendMicroActions(ActNameMicroDelta_s,AgentTck,StepNum) :-  
	m_out(proposedAction(StepNum,AgentTck,ActNameMicroDelta_s)),
	format("              sending actions ~w~n", [ActNameMicroDelta_s]).


processResponses([],_StepNum,_AgentTck,_ActionNameMulti,_Promised,[]).
processResponses([Res|Respns_s],StepNum,AgentTck,ActionNameMulti,Promised,[R|Responses]) :-
	processResponse(Res,StepNum,AgentTck,ActionNameMulti,Promised,R),  %*** SERVE Promised??? Deve essere per-agente o per-planner?
	processResponses(Respns_s,StepNum,AgentTck,ActionNameMulti,Promised,Responses).  %ora e' per tutti (e sempre [] ?)

%ora stampa... Nei primi tre casi si restituisce (ultimo argomento) la lista (PromisedActs)
%di azioni che l'agente si e' impegnato a compiere e i cambiamenti allo stato (DeltaInState)
%avvenuto nella transizione appena eseguita
processResponse(success(Act,DeltaInState), StepNum, Tck, _, PromisedActs, success(PromisedActs,Act,DeltaInState)) :-
	format("Agent ~w-~w got success in step ~w.~n", [Tck,Act,StepNum]).
processResponse(conflicting(Act,DeltaInState), StepNum, Tck, _, PromisedActs, conflicting(PromisedActs,Act,DeltaInState)) :-
	format("Agent ~w-~w got conflicting in step ~w.~n", [Tck,Act,StepNum]).
processResponse(subsumed(Act,DeltaInState), StepNum, Tck, _, PromisedActs, subsumed(PromisedActs,Act,DeltaInState)) :-
	format("Agent ~w-~w got subsumed in step ~w.~n", [Tck,Act,StepNum]).

% Le prossime due hanno senso solo se si usa conflictSolve_comm_client.pl. (ATTUALMENTE DA ULTIMARE, TODO...)
% Caso di conflitto di un solo agente. Non dovrebbe verificarsi mai....
processResponse(call_for_agreement(_-[T]), _StepNum, Tck, _, _OldReqToPlanner, 'UNEXP_ERROR') :- !,
	format(" ~n***** UNEXPECTED ERROR: Agent ~w got call_for_agreement between ONE agent ?!? ~w.~n~n", [Tck,[T]]).
% Caso di conflitto: supervisor richiede una coordinazione tra i confliggenti:
processResponse(call_for_agreement(MyActs-AllT_As_list), StepNum, MyTck, ActionNameMulti, _OldReqToPlanner, WhatsNext) :-
	format("Agent ~w got call_for_agreement during step ~w because of actions ~w~n all agents:~w~n", [MyTck,StepNum,MyActs,AllT_As_list]),
	%write('STEP ME client:processResponse (type an atom):'),read(_),nl,
	conflictResolution_client(AllT_As_list, StepNum, MyTck, MyActs, PromisedActs, Response),   %***TODO
	processResponse(Response,StepNum,MyTck,ActionNameMulti, PromisedActs, WhatsNext).

%TODO:  chiarire l'uso di   MyAct in   call_for_agreement(MyActs-AllT_As_list)  e di  ActionNameMulti
%VERIFICARE CHE  ActionNameMulti e' la multi azione da eseguire e MyActs e' la lista dei nomi di azioni "semplici" in ActionNameMulti ?!?!?






setInitialState(AgentName,Priority,AgentTck,InitialPairs,Response) :-
	m_rd(startSettingInitialState),
	format("Sending my name and priority, agentReady(AgentName,Priority)  : ~w~n", [agentReady(AgentName,Priority)]),
	m_out(agentReady(AgentName,Priority)),
	m_in(ticket(AgentName,AgentTck)),
%	format("Receiving myID, ticket(AgentName,AgentTck) : ~w~n", [ticket(AgentName,AgentTck)]),
% per perdere un po' di tempo:
%          ((AgentName==2) -> (write('WWWWWWWWWWWWAIT'),nl,sleep(15)); sleep(0)),
	format("Agent ~w got the ticket:~w~n", [AgentName, AgentTck]),
%	format("Spedisco initialState(AgentTck,InitialPairs)   con~n          AgentName   =~w~n",[AgentName]),
%	format("          AgentTck    =~w~n",[AgentTck]),
%	format("          InitialPairs=~w~n",[InitialPairs]),
	m_out(initialState(AgentTck,InitialPairs)),
%	format("      SPEDITO:~w~n",[initialState(AgentTck,InitialPairs)]),
	format("Agent ~w is waiting for the response...~n", [AgentTck]),
	m_in(initialStateResponse(AgentTck,Response)),
	format("Agent ~w  received response:~w~n", [AgentTck, Response]).


startConnection(Address,Port) :-
	format(" Client: Address      = ~w~n",[Address]),
	format("         Port         = ~w~n",[Port]),
	m_linda_client(Address:Port).

closeConnection(AgentName) :-
	format(" Client ~w ending. Closing connection...~n",[AgentName]),
	m_close_client.





