:- module(conflictSolve_comm_client, [conflictResolution_client/6]).

% conflictResolution_client/6  to be invoked by each client/agent
%  see below for the details...
% See also
% conflictResolution_supervisor/4  to be invoked by the supervisor (defined elsewhere)

:- use_module(library(lists)).
:- use_module(m_lindaclient).
:- use_module('plan_executor.pl', [getAcceptableRequest/2,get_on_X_Policies/3]).




%%%%%%%%%%%%%%
%TODO:
%%%%%%%%%%%%%%
%
% applyMyOption e' abbozzata, presumo si debba interfacciare meglio col planner
% ora ad esempio il fatto che si e' ricorso a ARBITRATION non viene comunicato ai planner... 
%
% applyMyOption(+Option, -MessageToSupervisor, -MessageToPlanner)
% per ogni agente,  MessageToPlannerA  deve contenere le indicazioni 
% per il pianificatore su cosa fare a seguito della contrattazione (es. ripianificare,
% oppure "la prossima azione deve essere NO-OP", etc
%%%%%%%%%%%%%%
%ECCO L'ATTUALE CODICE
%
%applyMyOption(+Option, -MessageToSupervisor, -MessageToPlanner)
%RETRY-OP per ora e' generato solo in risposta a subgoal richiesto da altri agenti
applyMyOption(forego, 'NO-OP', ['REPLAN']).  % ripianificare. 'REPLAN' 
applyMyOption(arbitration, 'ARBITRATE', ['ARBITRATE']).
applyMyOption(retry_after(T),  'NO-OP', ['EXEC-NO-OPs'(T1)]) :- T > 0, T1 is T-1. %eseguire T-1 azioni NO-OP: T conta ANCHE una NO-OP appena eseguita
applyMyOption(forego(Clist),  'ASK'(Clist), ['REPLAN'(Clist)]).
applyMyOption(retry_after(T,Clist), 'ASK'(Clist), ['EXEC-NO-OPs'(T1,Clist)]) :- T > 0, T1 is T-1.

%%%%%%%%%% END TODO %%%%%%%%%
%

%Mar 2011: scalo da una azione a multi azioni:
% conflictResolution_client(+AllT_As_list, +StepNum, +MyTck, +MyActs, -PromisedActs, -Response)
% AllT_As_list: lista di coppie Ticket-ListaActsNames di tutti gli agenti coinvolti nel conflitto
% MyActs: lista dei nomi delle azioni che MyTck intendeva eseguire
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% conflictResolution_client(+AllTckts, +StepNum, +MyTck, +ActionName, -PromisedActs, -Response),	(OLD,Mar11)
%
% AllTckts: lista dei ticket degli agenti coinvolti nel conflitto	(OLD,Mar11)
% StepNum: numero del passo in esecuzione
% MyTck: ticket dell'agente che invoca conflictResolution_client
% ActionName: azione che MyTck intendeva eseguire	(OLD,Mar11)
% PromisedActs: a seguito dell'interazione con altri agenti MyTck
%        potrebbe "promettere" di eseguire delle azioni. Questa e' tale lista.
% Response: esito della contrattazione (success, conflict, ...) relativamente a MyTck.
%        e' un termine del tipo success(ListaDiFluVal) dove ListaDiFluVal
%        raccoglie i cambiamenti dei flienti. Tale termine viene generato dal supervisor
%        una volta che la contrattazione e' terminata

% ROUND-ROBIN-based-communication
% Si procede cosi': gli agenti in AllTckts sono tutti in esecuzione
% contemporaneamente. Tuttavia, uno alla volta, nell'ordine
% in cui appaiono in AllTckts hanno ruolo attivo nella contrattazione.
% Chi ha ruolo attivo, esamina una delle opzioni on_conflict dichiarate
% nella definizione della azione ActionName. Se questa prevede decisione
% autonoma (es. forego, ...) la decisione viene comunicata a supervisor
% e si attiva il successivo agente nella lista. Se invece la opzione
% prevede un "provide Condizione", l'agente cerca tra gli altri agenti 
% (chiedendo ad uno ad uno) qualcuno che gli possa garantire Condizione.
% Se lo trova, comunica a supervisor la decisione corrispondente.
% Anche l'agente che ha accettato di garantire Condizione comunica la sua
% decisione. 
% Chi ha ruolo passivo attende o di essere attivato o una richiesta di "aiuto"
% nel realizzare una Condizione. In tal caso puo' accettare o meno.
% Se accetta eseguira' una azione che ATTUALMENTE DEVE essere la stessa
% ActionName (la sua) che aveva proposto all'inizio del passo di planning.
% (Si puo' "facilmente" prevedere che si esegua una azione diversa,
% in tal caso il supervisor dovra' verificare che la nuova azione non
% generi altri conflitti con le azioni di altri agenti (non necessariamente
% in AllTckts)).
% Terminato un turno (ovvero, quando tutti gli agenti in AllTckts hanno rivestito
% il ruolo attivo una volta), se vi sono ancora agenti che non hanno
% deciso, si comincia un nuovo turno (con solo gli agenti ancora "indecisi").
%
% Ogni agente attivo ottiene (tramite get_on_X_Policies/3) al turno i-esimo la
% i-esima opzione on_conflict dichiarata nella definizione della azione ActionName.
% (Quindi, tali opzioni sono processate nell'ordine in cui sono dichiarate
% nella action theory).
%
% Ogni agente che abbia comunicato a supervisor la sua decisione resta in attesa 
% di una tupla (actionResponse(StepNum,MyTck,Response)) che contiene
% l'esito del StepNum-esimo passo di planning (Response)


% conflictResolution_client(+AllT_As_list, +StepNum, +MyTck, +MyActs, -PromisedActs, -Response)
conflictResolution_client([MyTck-ActNames], _StepNum, MyTck, _MyActs, [], conflicting([])) :- !,
	format(" ~n*****WARNING CONFLICTRESOLUTION********: singoletto inconsistente!!?? ~w~n",[MyTck-ActNames]).


conflictResolution_client([FirstTck|Players], StepNum, MyTck-Acts, ActionName, PromisedActs, Response) :-
OPPURE
conflictResolution_client([FirstTck-FistActs|Players], StepNum, MyTck, MyActs????, PromisedActs, Response) :-
OPPURE
conflictResolution_client([FirstTck-FistActs|Players], StepNum, MyTck, MyMultiActCs, PromisedActs, Response) :-

	( (FirstTck == MyTck) -> m_out(turn(StepNum,1,MyTck,[FirstTck|Players])) ; true ),
	roundrobin(StepNum, MyTck, Acts, PromisedActs, Response).


roundrobin(StepNum, MyTck, Acts, PromisedActs,AgreementOutcome) :-
	m_in([ turn(StepNum,_AskRoundNum,MyTck,_StillPlayers),
	       proposal(StepNum,_AnsRoundNum,MyTck,_FromTck,_Request) ], Tuple),
	roundrobin2(Tuple,Acts,PromisedActs,AgreementOutcome).


roundrobin2(proposal(StepNum,AnsRoundNum,MyTck,FromTck,AskedSubGoal),_Acts,InfoToPlanner,Response) :-
	getAcceptableRequest(FromTck,AskedSubGoal), !, %accept proposal or not?
	m_out(answerToProposal(StepNum,AnsRoundNum,MyTck,FromTck,'ACCEPT')),
	m_out(proposedOption(StepNum,MyTck,'RETRY-OP')),
	m_in(actionResponse(StepNum,MyTck,Response)),
	%aggiungere 'subsumed'?
	(functor(Response,success,_) -> (InfoToPlanner=['RETRY-OP'(AskedSubGoal)]) ; (InfoToPlanner=[]) ).
%nella precedente, AskedSubGoal serve a comunicare planner che l'agente ha promesso di
%garantire AskedSubGoal
roundrobin2(proposal(StepNum,AnsRoundNum,MyTck,FromTck,_Clist),Acts,PromisedActs,Response) :- %proposal not accepted
	m_out(answerToProposal(StepNum,AnsRoundNum,MyTck,FromTck,'NOT-ACCEPT')),
	roundrobin(StepNum, MyTck, ActionName, PromisedActs,Response).
roundrobin2(turn(StepNum,AskRoundNum,MyTck,Players),Acts,PromisedActs,Response) :- %my AskRoundNum-th turn in round robin
	roundrobin2_aux(StepNum,AskRoundNum,MyTck,Players,Acts,PromisedActs_s,Response_s),

	IN SOSPESO: NON SERVE PER IL calcio
	
	myNBestOption(AskRoundNum,MyTck,ActionName,Option),
	applyMyOption(Option,  ToSupervisor, Acts),
	roundrobin3(ToSupervisor,StepNum,AskRoundNum,MyTck,Acts,Players,Acts,PromisedActs,Response).




roundrobin3('ASK'(Clist), StepNum,AskRoundNum,MyTck,ActionName,PlayingNow,Acts,PromisedActs,Response) :-
	( searchForPartner(PlayingNow,Partner,MyTck,StepNum,AskRoundNum,Clist)
		-> (
		    PromisedActs=Acts,
		    selectchk(Partner,PlayingNow,Rest1), %toglie la prima occorrenza di Partner
                                                         %Mar11: ok, per come def searchForPartner (che puo' selez. solo
                                                         %la prima occorrenza di un Tck
		    ( nextActor(MyTck, Rest1,NextTck,NewPlayers,AskRoundNum,AskRoundNum1) %prevedere piu' azioni per stesso MyTck
		    	-> m_out(turn(StepNum,AskRoundNum1,NextTck,NewPlayers)) ; true ),
		    m_in(actionResponse(StepNum,MyTck,Response))
	    	   )
		;  (
		    ( nextActorBis(MyTck, PlayingNow, NextTck, AskRoundNum,AskRoundNum1) %prevedere piu' azioni per stesso MyTck
		    -> (
			m_out(turn(StepNum,AskRoundNum1,NextTck,PlayingNow)),
			roundrobin(StepNum, MyTck, ActionName, PromisedActs, Response)
		       )
		    ; (  % MyTck is the only one, so none can provide Clist, but
			 % there might be further on_conflict options for MyTck,
			 % so MyTck activates ist next round...
			R is AskRoundNum+1,
			m_out(turn(StepNum,R,MyTck,PlayingNow)),
			roundrobin(StepNum, MyTck, ActionName, PromisedActs, Response)
		       )
		    )
		   )
	).
roundrobin3('NO-OP', StepNum,AskRoundNum,MyTck,_ActionName,Players,Acts,Acts,Response) :-
	m_out(proposedOption(StepNum,MyTck,'NO-OP')),
	( nextActor(MyTck, Players,NextTck,NewPlayers,AskRoundNum,AskRoundNum1) %prevedere piu' azioni per stesso MyTck
       		-> m_out(turn(StepNum,AskRoundNum1,NextTck,NewPlayers)) ; true ),
	m_in(actionResponse(StepNum,MyTck,Response)).
roundrobin3('ARBITRATE', StepNum,AskRoundNum,MyTck,_ActionName,Players,Acts,Acts,Response) :-
	m_out(proposedOption(StepNum,MyTck,'ARBITRATE')),
	( nextActor(MyTck,Players,NextTck,NewPlayers,AskRoundNum,AskRoundNum1) %prevedere/skippare piu' azioni per stesso MyTck
		-> m_out(turn(StepNum,AskRoundNum1,NextTck,NewPlayers)) ; true ),
	m_in(actionResponse(StepNum,MyTck,Response)).




searchForPartner([],_,_,_,_,_) :- !,fail.
searchForPartner([FromTck|Ts], Partner, FromTck, StepNum,AskRoundNum,Clist) :- !,
        % different from itself...
	searchForPartner(Ts, Partner, FromTck, StepNum,AskRoundNum,Clist).
searchForPartner([ToTck|Ts], Partner, FromTck, StepNum,AskRoundNum,Clist) :- !,
	m_out(proposal(StepNum,AskRoundNum,ToTck,FromTck,Clist)),
	m_in(answerToProposal(StepNum,AskRoundNum,ToTck,FromTck,Ans)),
	( (Ans=='ACCEPT')
	-> (
	    m_out(proposedOption(StepNum,FromTck,'NO-OP')),
	    Partner=ToTck
	   )
	;  (
   % XXX continua escludendo altre occorrenze di ToTck in Ts
	    searchForPartner(Ts, Partner, FromTck, StepNum,AskRoundNum,Clist)
	   )
	).


%select the element following Tck in the list Tckts (considered circular)
%NewTckts is Tckts without Tck, Round1 is Round+1 when Tck is the last one in Tckts.
%(assume that Tck occurs in Tckts)
%nextActor(Tck, [Tck], Next, NewTckts, Round, Round) :- !,fail.
nextActor(Tck, Tckts, Next, NewTckts, Round, Round) :-
	%format("1 nextActor(Tck, Tckts, Next, NewTckts, Round, Round) ~w~n",[nextActor(Tck,Tckts,Next,NewTckts,Round,Round)]),
	nextto(Tck, Next, Tckts),!,
	%format("2 nextActor(Tck, Tckts, Next, NewTckts, Round, Round) ~w~n",[nextActor(Tck,Tckts,Next,NewTckts,Round,Round)]),
	selectchk(Tck,Tckts,NewTckts).
nextActor(Tck, [Next,Tck2|Tckts], Next, NewTckts, Round, Round1) :- !,
	%format("3 nextActor(Tck, Tckts, Next, NewTckts, Round, Round) ~w~n",[nextActor(Tck,Tckts,Next,NewTckts,Round,Round)]),
	Round1 is Round+1,
	%format("4 nextActor(Tck, Tckts, Next, NewTckts, Round, Round) ~w~n",[nextActor(Tck,Tckts,Next,NewTckts,Round,Round)]),
	selectchk(Tck,[Tck2|Tckts],NewTckts).

%as the previous but do not remove Tck:
nextActorBis(Tck, Tckts, Next, Round, Round) :-
	nextto(Tck, Next, Tckts),!.
nextActorBis(_Tck, [Next,_Tck2|_Tckts], Next, Round, Round1) :- !,
	Round1 is Round+1.



%get_on_X_Policies/3 def in plan_executor.pl, supposta usare 
% un predicato on_X_Policies/3 definito dall'interfaccia di ogni planner. (PRO-BAAC).
myNBestOption(N,_Tck,ActionName,Option) :-
	get_on_X_Policies(ActionName, OnCPs, _OnFPs),!,
	(nth1(N,OnCPs,Option) -> true ; (Option=forego)).
	% the default (if no (more) option exists) is failure della azione






