:- module(plan_executor, [auto_run_agent/1,getAcceptableRequest/2,get_on_X_Policies/3]).

:- use_module(library(system)).
:- use_module(library(lists)).
:- use_module('settings.pl').
:- use_module('client.pl').

%since auto_run_agent/1 loads both the planner and the action theory,
% auto_run_agent/1 HAS to be called first, before any other call
% to getAcceptableRequest/2, get_on_X_Policies/3,...

% AgentIndex --> AgentName --> AgentTck&Priority
auto_run_agent(AgentIndex) :-
    theorySourceFile(AgentIndex,AgentName,Filename,PlannerModule),
    format("~n*********~nLoading the planner ~w~n*********~n", [PlannerModule]),
    loadThePlanner(PlannerModule),
    %write('STEP ME (type an atom):'),read(_),
    format("~n*********~nLoading the action theory ~w~n*********~n", [Filename]),
    loadTheTheory(Filename), 
    format("~n*********~nEstablishing connection with the supervisor~n*********~n", []),
    getConnectionData(Address,Port),
    startConnection(Address,Port),
    format("~n*********~nRunning client/planner~n*********~n", []),
    execPlan(AgentName),
    format("Client ending. Closing connection...~n",[]),
    closeConnection(AgentName).

auto_run_agent(AgentIndex) :-
    format("FAILURE in loading the planner or the action theory for agent number:~w~n", [AgentIndex]).

loadThePlanner(Filename) :- 
    once(use_module(Filename, [load_theory/1,mkInitialState/1, meta_planner_multi/5, meta_plannerPrefixed_multi/8, 
                                   getAgentPriority/1,on_X_Policies/3,setFluentsInState/2])).

loadTheTheory(Filename) :- 
    once(load_theory(Filename)).


getConnectionData(Address,Port) :-
    temporaryFileName(FileName),
    see(FileName),
    read(connectionAddress(Address,Port)),
    seen.


execPlan(AgentName) :-
% per perdere un po' di tempo:
    format("Waiting 3 sec... for houskeeping.... ~n",[]),
    sleep(3),
    %write('STEP ME (type an atom):'),read(_),
    format("~n*********~nSetting initial state ~n*********~n", []),
    mkInitialState(InitialState),!,
    initializeState(AgentName,InitialState,AgentTck),!,
    format("~n*********~nSearching for the (shortest) initial plan ~n*********~n", []),
    %ERA mkFirstPlan(AgentName, ActionDeltaLists, DeltasTrajectory, _StateSequence),
    mkFirstPlan(AgentName, MultiActionEffect_Lists, EffectsTrajectory, _StateSequence),
    format("AgentTck ~w: Effects: ~w~n", [AgentTck,EffectsTrajectory]),
    %format("~n*********~nExecuting the plan ~n*********~n", []),
    performSteps(AgentTck, MultiActionEffect_Lists, EffectsTrajectory),
    endingStuff(AgentTck).
execPlan(AgentName) :-
    format("AgentName ~w: plan execution failure.~n", [AgentName]).

mkFirstPlan(_AgentName,  MultiActionEffect_Lists, EffectsTrajectory, StateSequence) :-
    once(maxNumberOfSteps(Steps)),
    meta_planner_multi(1,Steps, MultiActionEffect_Lists, EffectsTrajectory, StateSequence). %***




initializeState(AgentName,InitialPairs,AgentTck) :-
    getAgentPriority(Priority),
    setInitialState(AgentName,Priority,AgentTck,InitialPairs, Ackno), !,
    ( (Ackno = incoherent(Conflicts))
        -> ( format("AgentTck ~w: INCOHERENT GLOBAL INITIAL STATE:~w~n", [AgentTck, Conflicts]),
             !, fail )
        ;    format("AgentTck ~w: setInitialState succeeded~n", [AgentTck])
    ).


%PlanMulti e' una lista di liste Li. Ognuna delle liste Li e' la lista delle azioni compiute
%in una state transition (il planner potrebbe essere multi-agente, altrimenti Li e' singleton)
%Ogni Li e' una lista di oggetti della forma ActName-AffectefFluents, dove ActName e' l'azione
%e AffectefFluents e' una lista dei fluenti influenzati da ActName nella transizione
%Se il planner e' mono-agente, Li e' singoletto che puo' essere anche della forma [ActionName].
performSteps(AgentTck, PlanMulti, DeltasTrajectory) :- 
    performSteps(AgentTck, 1, [], PlanMulti, DeltasTrajectory). 
%performSteps(+AgentTck, N, +PlanDone, +PlanToDo, +DeltasTrajectoryToDo) :- 
% PlanToDo and PlanDone are lists of lists of action names (to handle multi-agents) s.t. the
% entire PlanMulti is PlanMulti=append(rev(PlanDoneMulti),PlanToDoMulti)
% N is the number of the current step (N should be length(PlanDoneMulti)+1
% DeltasTrajectory is the list of lists of pairs [F,V]. Such pairs encode the changes in PlanToDoMulti
% from current state (FromState) to the following one (ToState)
% by effect of the (multi-)action execution.
% (such effects are "expected": they will be realized if the action(s) 
% execution succeeds... otherwise the plan could/should be revised.
performSteps(AgentTck, _, RevPlanDone, [], []) :-
    terminalStep(AgentTck),
    format("Agent ~w :performSteps: plan execution ended.~n", [AgentTck]),
    format("          reverse executed plan: ~w~n", [RevPlanDone]).
performSteps(AgentTck, NumCurrentTransition, RevPlanDone, [ActionNamesMulti|As], [Delta|Deltas]) :- 
    format("Agent ~w :performSteps:attempting ~w-th transition... ~n", [AgentTck,NumCurrentTransition]),
    format("              PlanToDo=~w~n", [[ActionNamesMulti|As]]),
    %write('plan_executor:performSteps: STEP ME (type an atom):'),read(_),
% itera esecuzioni di passi in sincronia con le richieste del supervisore, 
% fino a quando quest'ultimo ne richiede o fino a quando l'agente decide
% autonomamente di terminare
        performOneStep(AgentTck,ActionNamesMulti,Delta,NextActsReqToPlanner_Response_list), %***QUELLO IN client.pl 
    %write('plan_executor:performSteps: STEP ME (type an atom):'),read(_),
    format("Agent ~w :performSteps attempted one step. The response is:~w~n", [AgentTck,NextActsReqToPlanner_Response_list]),
    handleOutcomes(NextActsReqToPlanner_Response_list, AgentTck, NumCurrentTransition, RevPlanDone, [ActionNamesMulti|As], [Delta|Deltas]). 


handleOutcomes(NextActsReqToPlanner_Response_list, AgentTck, StepNum, RevPlanDone, [_LastAsMulti|OldNextAsMulti], OldDeltas) :-
        collectJustDoneAndTodo(NextActsReqToPlanner_Response_list, RealDelta, NextStepsOfPromisedActs, ThisStepJustDoneActs, IntermediateConstraints), 
    setIntermediateConstraints(IntermediateConstraints),
    revisePlan(StepNum, NextStepsOfPromisedActs, AgentTck, RealDelta, RevPlanDone, [ThisStepJustDoneActs|OldNextAsMulti], OldDeltas, NewRevPlanDone, NewNextActionsMulti, NewDeltas),
    N1 is StepNum +1,
    performSteps(AgentTck, N1, NewRevPlanDone, NewNextActionsMulti, NewDeltas).


%TODO: da implementare (serve per round-robin e simili strategie in cui gli agenti assumono/promettono constraint)
setIntermediateConstraints(_IntermediateConstraints).


%collectJustDoneAndTodo(NextActsReqToPlanner_Response_list, RealDelta, NextStepsOfPromisedActs, ThisStepJustDoneActs, IntermediateConstraints)
collectJustDoneAndTodo([],[],[],[],[]) :- !.
collectJustDoneAndTodo(Response_list,[],[],[],[]) :-  %%%,[],_,_):-
    memberchk(interactionEnded, Response_list),!.
collectJustDoneAndTodo(Response_list,RealDelta,NReqActs_s,L2,IntermediateConstraints) :-
        member(Resp,Response_list),!, Resp =.. [_SuCoOrSub,_,_Act,RealDelta|_UsuallyEmpty],
    collectJustDoneAndTodo_bis(Response_list,NReqActs_s,L2,IntermediateConstraints).

collectJustDoneAndTodo_bis([],[],[],[]).
collectJustDoneAndTodo_bis([success(NextReqActsKey,AgsActMulti,_DeltaInState)|Rest],PromisedActsemiplan,[AgsActMulti|L2], IntermediateConstraints) :- !,
        mknextactions(NextReqActsKey,AgsActMulti,NReqActs_semiplan,IntermediateConstraints1),
        collectJustDoneAndTodo_bis(Rest,NReqActs_semiplan1,L2,IntermediateConstraints2),   %NReqActs_semiplan describes a list of partial transitions_multi
        append(IntermediateConstraints1, IntermediateConstraints2, IntermediateConstraints),
        mergeParallel(NReqActs_semiplan, NReqActs_semiplan1, PromisedActsemiplan).
collectJustDoneAndTodo_bis([conflicting(NextReqActsKey,AgsActMulti,_DeltaInState)|Rest],PromisedActsemiplan,L, IntermediateConstraints) :- !,
        mknextactions(NextReqActsKey,AgsActMulti,NReqActs_semiplan,IntermediateConstraints1),
        collectJustDoneAndTodo_bis(Rest,NReqActs_semiplan1,L2,IntermediateConstraints2),
        append(IntermediateConstraints1, IntermediateConstraints2, IntermediateConstraints),
        AgsActMulti = act(AgsList,_Act),
        findall(act([OneAg],'NO-OP'), member(OneAg,AgsList), Pref),
        append(Pref,L2,L),
        mergeParallel(NReqActs_semiplan, NReqActs_semiplan1, PromisedActsemiplan).
collectJustDoneAndTodo_bis([subsumed(NextReqActsKey,AgsActMulti,_DeltaInState)|Rest],PromisedActsemiplan,[AgsActMulti|L2], IntermediateConstraints) :- !,
        mknextactions(NextReqActsKey,AgsActMulti,NReqActs_semiplan,IntermediateConstraints1),
        collectJustDoneAndTodo_bis(Rest,NReqActs_semiplan1,L2,IntermediateConstraints2),
        append(IntermediateConstraints1, IntermediateConstraints2, IntermediateConstraints),
        mergeParallel(NReqActs_semiplan, NReqActs_semiplan1, PromisedActsemiplan).



%mknextactions(NextReqActsKey,LastAsMulti,NReqActs_listoflists,IntermediateConstraints)
%TODO mknextactions da verificare/completare a seguito delle azioni multiagenti (act(AgentList,Actname)):
mknextactions([],_,[],[]).
mknextactions(['RETRY-OP'(Clist)|Rest],LastAsMulti,NReqActs,Cs) :- !,
    mknextactions(Rest,LastAsMulti,NReqActs,Cs1),
    append(Clist,Cs1,Cs).
mknextactions(['EXEC-NO-OPs'(0)|Rest], LastAsMulti,NReqActs,Cs) :- !, %superflua...
    mknextactions(Rest,LastAsMulti,NReqActs,Cs).
mknextactions(['EXEC-NO-OPs'(1)|Rest], LastAsMulti,NReqActs,Cs) :- !,  %per non fallire
    mknextactions(Rest,LastAsMulti,NReqActs,Cs).
mknextactions(['EXEC-NO-OPs'(0,_ClistOfAnother)|Rest], LastAsMulti,NReqActs,Cs) :- !, %superflua...
    mknextactions(Rest,LastAsMulti,NReqActs,Cs).
mknextactions(['EXEC-NO-OPs'(1,_ClistOfAnother)|Rest], LastAsMulti,NReqActs,Cs) :- !,  %per non fallire
    mknextactions(Rest,LastAsMulti,NReqActs,Cs).
mknextactions(['EXEC-NO-OPs'(T)|Rest],LastAsMulti,NReqActs,Cs) :- T>1, !, 
        T1 is T - 1,
        buildNnoopsList(T1,LastAsMulti,Noops),
    mknextactions(Rest,LastAsMulti,NReqActs2,Cs),
        mergeParallel(Noops,NReqActs2,NReqActs).
mknextactions(['EXEC-NO-OPs'(T,_ClistOfAnother)|Rest],LastAsMulti,NReqActs,Cs) :- T>1, !, 
        T1 is T - 1,
        buildNnoopsList(T1,LastAsMulti,Noops),
    mknextactions(Rest,LastAsMulti,NReqActs2,Cs),
        mergeParallel(Noops,NReqActs2,NReqActs).
%TODO: da completare le seguenti, ora in versione "essenziale": causano ripianificazione "pura"
mknextactions(['ARBITRATE'|_Rest],_,[],[]) :- !.
mknextactions(['REPLAN'|_Rest],_,[],[]) :- !.
mknextactions(['REPLAN'(_Clist)|_Rest],_,[],[]) :- !.
%MEMO-LEGGENDA:
%      'REPLAN' = necessario ripianificare da zero
%      'ARBITRATE' = e' avvenuto arbitraggio (solo chi li ha chiesto lo sa....)
%      'EXEC-NO-OPs'(T) = promessa di eseguire T azioni NO-OP (considerando che UNA e' stata appena eseguita)
%      'REPLAN'(Clist) =  necessario ripianificare, considerando che un altro agente ha promesso
%                         di garantire Clist
%      'EXEC-NO-OPs'(T,Clist) = promessa di eseguire T azioni NO-OP (considerando che UNA e' stata appena eseguita),
%                               considerando che un altro agente ha promesso di garantire Clist 
%      'RETRY-OP'(Clist) = azione eseguita con successo e promessa di garantire Clist 
%  PER ORA: tutte le Clist sono/saranno imposte nel PROSSIMO stato

mergeParallel([],Lists2,Lists2).
mergeParallel([L|Lists1],[],[L|Lists1]).
mergeParallel([L1|Lists1],[L2|Lists2],[M|Merged]) :-
       append(L1,L2,M),
       mergeParallel(Lists1,Lists2,Merged).
        

buildNnoopsList(0,_,[]) :- !.
buildNnoopsList(N,act(AgsList,Act),[Pref|Rest]) :- N > 0,
        findall(act([OneAg],'NO-OP'), member(OneAg,AgsList), Pref),
        N1 is N-1,
        buildNnoopsList(N1,act(AgsList,Act),Rest).


%%%%%%%%%%%%%%%
%TODO: da aggiornare se introduciamo riferimenti in avanti:
updateLocalState(ListOfFluVal,StepNum) :-
    StateNum is StepNum,  %+1,
    setFluentsInState(ListOfFluVal,StateNum).
%%%%%%%%%%%%%%%




%%%%%%%%%%%%
%%%%%%%%%%%% NEL CHIAMARE meta_plannerPrefixed_multi, IL PREFISSO DEVE PREVEDERE ANCHE LE AZIONI DEL PASSATO
%%%%%%%%%%%% (non e' detto infatti che il planner ricalcoli sempre lo stesso prefisso
%%%%%%%%%%%%  Quindi devo portar dietro tutto il Plan e non solo i passi futuri...)
%%%%%%%%%%%%
revisePlan(ThisStepNum, NextReqActs_s, AgentTck, RealDelta, RevPlanDone, [LastAsMulti|_OldNextAsMulti], _OldDeltas, NewRevPlanDone, NewNextActionsMulti, NewDeltas) :-
    format("Agent ~w :start revising plan.~n", [AgentTck]),
    format("      AgentTck    = ~w ~n", [AgentTck]),
    format("      ThisStepNum = ~w ~n", [ThisStepNum]),
    format("      NextReqActs = ~w ~n", [NextReqActs_s]),
    format("      RealDelta   = ~w ~n", [RealDelta]),
    format("~n*********~nLocal knowledge update because of exogenous effects and constraints~n*********~n", []),
        updateLocalState(RealDelta,ThisStepNum),
        maxNumberOfSteps(MSteps),
        length(RevPlanDone,L1), length(NextReqActs_s,L2), AtLeastSteps is L1+L2+1, %OSS: AtLeastSteps>=ThisStepNum
        !,
        format("~n*********~nValidating/Modifying the plan because of exogenous effects and constraints~n*********~n", []),
        meta_plannerPrefixed_multi(AtLeastSteps, MSteps,[LastAsMulti|RevPlanDone], NextReqActs_s, ThisStepNum, NewRevPlanDone, NewNextActionsMulti, NewDeltas),
        !,
        format("Agent ~w : re-planning done.~n", [AgentTck]),
        format("      ThisStepNum    = ~w ~n", [ThisStepNum]),
        format("      NewNextActions = ~w ~n", [NewNextActionsMulti]),
        format("      NewDeltas      = ~w ~n", [NewDeltas]).


% revisePlan(...)
% NextReqActs = list of actions to be executed (promised)
% AgentTck = the agent
% RealDelta = the changes in state really caused by concomitant actions' execution
% OldNextAsMulti = the rest of the plan (multi-steps) as computed before the current state transition
% OldDeltas = the rest of the deltas in the trajectory as computed before the current state transition
% NewNextActionsMulti, NewNextDeltas = OldNextAsMulti and OldDeltas revised by effect of the Response
%
%-------------- Vecchie note IN PARTE superate -----------------------
% Da sostituire con un predicato esportato da sicsplan.pl (manca!)
% Alcune necessita':
% 1) Nel caso Response=success  la esecuzione della azione e' riuscita. Occorre controllare
%    che si possa continuare con il piano precedente (codificato da OldNextAs e OldDeltas).
%    Infatti gli effetti delle azioni degli altri agenti (i cambiamenti allo stato sono
%    in RealDelta (lista di [F,V])) potrebbero non piu' supportare le azioni in OldNextAs.
%    Eventuale ripianificazione fornisce il nuovo piano: NewNextActions e NewNextDeltas (dal
%    prossimo passo in avanti).
%
% 2) Nel caso _Response=conflicting  la esecuzione della azione e' fallita. Occorre ripianificare
%    i cambiamenti dello stato dovuti agli altri agenti sono RealDelta
%    In caso di fallimento dell'azione, revisePlan dovrebbe tener conto delle opzioni "on_failure"
%    descritte nella action theory, per decidere se e come ripianificare.
%
%    OSS: al momento il supervisor attende da tutti gli agenti MaxStep azioni da eseguire!!!
%    Sarebbe da prevedere il caso che un agente termini prima (perche' raggiunge il suo goal
%    o perche' abbandona). Per ora assumo che tutti gli agenti che terminano prima di MaxStep passi
%    continui comunque a eseguire NO-OP fino a giungere a MaxStep passi.
%    
% 3) Tenere presente che a seguito della gestione di on_conflict l'agente potrebbe aver "promesso" di eseguire
%    una lista di azioni (solitamente e' una lista di NO-OP, originata da retry_after, ma in futuro
%    potremmo decidere di arricchire....)
%    Notare che le azioni promesse potrebbero non venir effettivamente eseguite da chi le ha promesse
%    perche' per effetto di altri agenti potrebbero non  essere piu' eseguibili, oppure
%    perche' chi ha promesso non mantiene. Relativamente questo ultimo caso si potrebbe pensare
%    a introdurre in futuro la nozione di trust: un agente si fida delle promesse di un altro
%    se in passato quest'ultimo ha mantenuto le promesse... etc
%
%    La lista NextReqActs contiene sia le azioni promesse che delle info sulle "cause" di tali promesse
%    Termini possibili nella lista sono descritti anche nel terzo argomento di applyMyOption/3
%    (vedi conflictSolve_comm_client.pl) :
%      'REPLAN' = necessario ripianificare
%      'ARBITRATE' = e' avvenuto arbitraggio (solo chi li ha chiesto lo sa....)
%      'EXEC-NO-OPs'(T) = promessa di eseguire T azioni NO-OP contando ANCHE la azione appena eseguita
%      'REPLAN'(Clist) =  necessario ripianificare, considerando che un altro agente ha promesso
%                         di garantire Clist
%      'EXEC-NO-OPs'(T,Clist) = promessa di eseguire T azioni NO-OP contando ANCHE la azione appena
%                               eseguita, considerando che un altro agente ha promesso di garantire Clist 
%      'RETRY-OP'(Clist) = azione eseguita con successo e promessa di garantire Clist 






% predicato che saluta e chiude la baracca....
% per ora non fa niente, ma potrebbe essere il caso di stampare qualcosa, de-allocare risorse, ... BOH!?!
endingStuff(_AgentTck).




%%%%%%%%%%%%%%
%TODO-->FATTO: getAcceptableRequest e' provvisoria, dovrebbe interagire con 
% un predicato  acceptableRequest/2  definito nel planner (e.g., in sicsplan.pl) per
% valutare se una condizione possa essere garantita dalla/e futura/e azione/i
% (in tal caso l'agente puo' accettare una proposal)

% A seguito di una richiesta di garantire la condizione Clist,
% l'agente deve decidere se accettare di aggiungere Clist come sotto-goal al suo goal
% (o verificare che ne fa gia' parte).
% Qui si potrebbe/dovrebbe comportare ri-pianificazione.
%  Per ora me la cavo con "ask user" ;)
getAcceptableRequest(ticketToRide(AgentName,_Pri),RequiredCond):- 
        format("~n****Checking acceptability of requirement from agent ~w to ensure condition ~w...",[AgentName,RequiredCond]),
    acceptableRequest(AgentName,RequiredCond),!,
        format("...YES!~n",[]).
getAcceptableRequest(_,_):- 
        format("...NO!~n",[]),!,fail.

%
%%%%%%%%%% END TODO %%%%%%%%%

get_on_X_Policies(ActionName, OnCPs, OnFPs) :-
    on_X_Policies(ActionName, OnCPs, OnFPs). 



