%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  Primo merging grezzo del codice del planner BMV con l'infrastruttura
%  distribuita su Linda:
%  Preso SICSplanMV09.txt su rete (al 30 marzo 2010)
%  *) semplificato il meno possibile (no static, ...
%  *) aggiunta la sintassi "()-free" da BMAP ma non curandosi
%  della parte static, global-constraint, concorrenza, costi,...
%  *) quindi ACCETTA ancora constraint complessi in input, ma "ASSUMIAMO"
%  di usare solo constraint del tipo  Fluent = Valore,
%  dove Valore e' univocamente determinato al momento in cui viene
%  usato (= passato allo spazio delle tuple")
%
%  Aggiunta sintassi per comunicazione e gestione conflitti:
%   "action Act"  puo' essere grossomodo:
%   action Act  ( on_conflict  <onconflict>  [provided C] )* 
%               ( on_failure  <onfailure>  )*
%   dove     <onconflict> ::=  retry_after T [provided C] 
%                             | forego [provided C]
%                             | arbitration
%            <onfailure> ::=   retry_after T [if C]
%                            | replan [if C] [add_goal C]
%                            | fail [if C]
%  default:   action Act on_failure replan
%  (in caso di piu' opzioni alternative si puo' scegliere la prima o l'ultima, o...
%   i "C" possono anche essere liste: C_1,C_2,... 
%   T e' un numero intero positivo)
%
%  conflict: indica che in un passo ci sono piu' azioni incompatibili.
%            la soluzione del conflitto avviene durante il passo e
%            determina un sottoinsieme delle azioni da eseguire
%            (compatibili, ovviamente). Questa fase puo' comportare
%            dialogo/contrattazione tra agenti e avviene durante
%            la transizione di stato.
%  failure:  indica che nella transizione di stato l'agente non
%            hanno potuto eseguire l'azione (es. quelle escluse 
%            dalla contrattazione fatta per risolvere un conflitto).
%            (Quindi in questo passo e' come se avesse eseguito no-op.)
%            Failure si concretizza dopo la transizione e comporta
%            solitamente ripianificazione (vedi sotto).
%
%  Semantica intuitiva: Siano a_1...a_n agenti che
%  provano ad eseguire le azioni  z_1...z_n al passo l.
%  Supp. che a_1...a_k siano in conflitto (i.e. gli
%  effetti di z_1...z_k sono incoerenti).
%  Si apre una contrattazione tra a_1...a_k in cui :
% -se (almeno) un "on_conflict arbitration" e' presente
%  allora si invoca una arbitrato da parte del supervisor 
%  che decidera' con qualche criterio. Altrimenti,
% -se z_i e' def con  "on_conflict forego provided C" 
%  allora   a_i rinuncia a eseguire z_i se qualcuno degli altri a_j
%  esegue una azione (in questo step) che garantisce C tra gli effetti;
% -se k_i e' def con  "on_conflict retry_after T provided C",
%  allora se qualcuno degli altri a_j esegue una azione (in questo step)
%  che garantisce C tra gli effetti, a_i  eseguira' T volte no-op e poi
%  ritentera' a_i;
% -se "provided C" manca si assume sia "true"
% -se nessuna delle due opzioni e' praticabile, (o in assenza della
%  loro definizione (il default)) si ha failure della azione.
%
%  In caso di failure della azione (se l'azione e' stata inibita o 
%  la gestione di on_conflict ha dato esito negativo) si applica
%  la politica descritta da on_failure:
%  - retry_after T [if C]:  valuta C (assente=true) e in caso
%  esegue T no-op e poi riprova la azione fallita;
%  - replan [if C1] [add_goal C2]:  valuta C1 (assente=true) e in caso
%  esegue, dal passo successivo la ripianificazione aggiungendo C2 al
%  goal attuale (assente=true)
%  - fail [if C1]  corrisponde a   "replan [if C1]  add_goal 0=1"
%  quindi il planner rinuncia e dichiara il fallimento nel
%  raggiungimento del suo goal.
%  - se nessuna delle "retry_after/replan" e' applicabile (considerate
%  nell'ordine in cui occorrono) e non c'e' una "fail" applicabile,
%  allora si procede come se ci fosse "replan if true"
%
%  Accettazione della contrattazione:
%  In fase di risoluzione del conflitto ogni agente coinvolto,
%  qualora la "on_conflict" preveda un "provided C", comunica
%  C agli altri confliggenti e verifica se uno di questi accetti di
%  compiere una azione che 
%  - garantisca C
%  - e sia compatibile con le azioni eseguite dai non-confliggenti.
%  In tal caso entrambi "escono dalla contrattazione".
%  L'iter avviene agente per agente, in "mutua esclusione", 
%  Potranno esserci agenti che restano "insoddisfatti" per i quali si
%  avra' "failure".
%
% GENERALIZZAZIONE INNOCUA (spero....) la condizione C delle
% opzioni  "on_conflict" puo' essere di piena generalita' (non necessariamente
% congiunzione di "F eq V") in quanto non viene utilizzata per
% modificate lo spazio delle tuple. Viene solo testata (predicato
% acceptableRequest (def nel planner e invocato, ad es., 
%   da getAcceptableRequest def in plan_executor, a sua volta 
%   usato in conflictSolve_comm_client.pl) e
% passata al pianificatore che la aggiungera' al goal/constraint/...
% Puo' quindi contenere riferimenti temporali in avanti.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%






%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- module(sicsplan, [load_theory/1,
                     mkInitialState/1,
		     %Non usato in plan_executor.pl      sicsplan/3,
                     meta_planner/4,
                     meta_planner_multi/5,
		     %Non usato in plan_executor.pl      sicsplanPrefixed/7,
                     meta_plannerPrefixed/8,
                     meta_plannerPrefixed_multi/8,
                     getAgentPriority/1,
                     on_X_Policies/3,
                     setFluentsInState/2,
                     acceptableRequest/2,  %(related to the next one)
                     thisAgentHelps_DNF/2]).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% INTERFACE PREDICATES:
% (plan_executor.pl expects them to be exported by all
% modules that implement a planner.

%load_theory(+Filename) loads a theory from Filename

%mkInitialState(-ListOfPairsFV)
% retrieves the initial (local) state as defined in the theory
% BEWARE: the state has to be determined, i.e. a list of
%         constraint of the form "fluent eq value"
% example:  mkInitialState([[a,3],[b,2],[c1,14],[d1,15]])

%sicsplan(+N,-ActionList,-StateList)
% given the lenght N of the plan sought for, succedes by
% instantiating ActionList to the list of N actions names and StateList 
% to the list of N+1 states (each state is a list of pairs [fluent,value])

%meta_planner(+StartLen,+MaxLen,-ActionList,-StateList)
% uses sicsplan/3 to search for the shortest plan
% with length between StartLen and MaxLen (inclusive)
%
%meta_planner_multi(+StartLen,+MaxLen,-ActionListMulti,-StateList)
% similar to meta_planner, except that ActionListMulti is a list of lists
% (actually, since this planner is a single-agent planner, each list
% is a singleton) of the form [...,ActionName-Delta,...]
% where ActionName performed in the step and Delta is the list of changes
% in the state caused by ActionName. Each change is a pair [Fluent,NewValue]



% versione con un prefisso di azioni predeterminato
%sicsplanPrefixed(+MaxSteps,+RevPlanDone,+PlanPrefixTodo,+ThisStepNum,-NewRevPlanDone,-NextActionList,-NextDeltas)
% succedes if NextActionList/NextDeltas describe the final part of a plan
% being at most MaxSteps long, such that the next step is the ThisStepNum-th,
% the plan starts with RevPlanDone and continues with PlanPrefixTodo
 
%meta_plannerPrefixed(+StartLen,+MaxLen,+RevPlanDone,+PlanPrefixTodo,+ThisStepNum,-NewRevPlanDone,-ActionList,-StateList)
% meta_plannerPrefixed is for sicsplanPrefixed what meta_planner is for sicsplan

%meta_plannerPrefixed_multi(+StartLen,+MaxLen,+RevPlanDoneMulti,+PlanPrefixTodoMulti,+ThisStepNum,-NewRevPlanDoneMulti,-ActionListMulti,-StateList)
% meta_plannerPrefixed_multi is for meta_plannerPrefixed what meta_planner_multi is for meta_planner

%getAgentPriority(-Pri)
% retrieves the priority of the agent(s)
% (a natural number, 0 is the highest)

%on_X_Policies(+ActionName,-OnCPs,-OnFPs)
% retrieves two lists of options for the action ActionName
% OnCPs is the list of policies of the kind on_conflict
% OnFPs is the list of policies of the kind on_failure
% the order is the one declared in the action theory

%setFluentsInState(+FluentValList,+StateNum)
% imposes constraints of the form Fluent eq Val
% in the StateNum-th state, for each pair [Fluent,Val]
% in FluentValList 
% (used to propagate effect of exogenous actions
% (i.e., effects of other agents' actions)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


:-use_module(library(clpfd)).
:-use_module(library(lists)).
%:-prolog_flag(unknown,_,fail).
:-prolog_flag(redefine_warnings,_,off).
:-prolog_flag(discontiguous_warnings,_,off).
:-use_module(library(terms)).




%%% In case no STATIC CAUSAL LAWS (caused) are defined in the ACT DES, add:
%%% Or no further constraints are defined, add
:- dynamic(caused/2).
:- dynamic(action_cost/2).
%:- dynamic(max_plan_cost/1).
:- dynamic(state_cost/1).
%:- dynamic(goal_cost/1).
:- dynamic(cost_constraints/1).
:- dynamic(holds/2).
:- dynamic(exogenous/2).
:- dynamic(always/1).
:- dynamic(time_constraint/1).
:- dynamic(action_bound/1).

%DAtxt:
:- dynamic(priority/1).
:- dynamic(agent/1).      :- dynamic(fluent/2).           :- dynamic(fluent/3).
:- dynamic(causes/3). % BMAP vs BMV
:- dynamic(causes/2).     :- dynamic(executable/3).       :- dynamic(action/2).
                                                          :- dynamic(action_cost/3).
                          :- dynamic(goal_cost/1).        :- dynamic(plan_cost/1).
:- dynamic(initially/1).  :- dynamic(goal/1).


:- dynamic(on_conflict/2).  :- dynamic(on_failure/2).


%%% %%% %%% %%% %%% %%% %%% %%% %%% %%% %%% %%%
%%% Infix operators for fluent constraints  %%%
%%% %%% %%% %%% %%% %%% %%% %%% %%% %%% %%% %%%

:-op(600,xfx,eq).
:-op(600,xfx,neq).
:-op(600,xfx,leq).
:-op(600,xfx,geq).
:-op(600,xfx,lt).
:-op(600,xfx,gt).
:-op(500,xfx,@).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% pretty syntax stuff:
:-dynamic(fluent/1).  :-dynamic(executable/1).  :-dynamic(costs/2).
:-dynamic(action/1).
:-dynamic(causes/1).
:-dynamic(causes/2). %BMAP vs BMV
:-op(1120,xfx,valued).
  :-op(1120,xfx,add_goal).
  :-op(1120,fx,retry_after).
%useless  :-op(.,.,replan).
:-op(1125,xfx,if).
%:-op(1125,xfx,costs).
:-op(1130,fx,executable).
:-op(1130,xfx,know).
:-op(1130,fx,causes).
:-op(1130,xfx,causes). %BMAP vs BMV
%:-op(1130,xfx,caused).
:-op(1130,xfx,costs).
:-op(1130,fx,state_cost).
:-op(1130,fx,plan_cost).
:-op(1130,fx,goal_cost).
:-op(1130,fx,initially).
:-op(1130,fx,goal).
:-op(1134,fx,priority).
:-op(1134,xfy,priority).
:-op(1135,fx,agent).
:-op(1135,fx,action).
:-op(1135,fx,fluent).
:-op(1135,xfx,holds).
:-op(1135,fx,time_constraint).
:-op(1135,xfx,provided).
:-op(1150,xfy,on_conflict).
:-op(1150,xfy,on_failure).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


load_theory(Filename) :-
	compile(Filename).


%mkInitialState(-ListOfPairsFV).
% example:  mkInitialState([[a,3],[b,2],[c1,14],[d1,15]]).
% IMPORTANTE: si assume che gli initially siano della forma "fluente eq valore"
mkInitialState(ListOfPairsFV) :-
	findall([FE1,OP,FE2],(orig_initially(C),C=..[OP,FE1,FE2]), InitFC),
	checkSimpleSyntax(InitFC,ListOfPairsFV).

checkSimpleSyntax([],[]).
checkSimpleSyntax([[FE1,eq,FE2]|InitFC],[[FE1,FE2]|ListOfPairsFV]) :-
	(fluent(F,_MinF,_MaxF) ; fluent(F,_Dom)),
	number(FE2), !,
	checkSimpleSyntax(InitFC,ListOfPairsFV).
checkSimpleSyntax([[FE1,OP,FE2]|InitFC],ListOfPairsFV) :-
	format("UNEXPECTED syntax of initial state constraint: ~w. Skipping it.~n", [[FE1,OP,FE2]]),
	checkSimpleSyntax(InitFC,ListOfPairsFV).




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%     pretty syntax digestion          %%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%UNPRETTY AGENT
% agent A  ==>  agent(A)   et   agentName(A)
% non usato  agentName(A) :- agent(A).
% non usato  getAgentName(Name) :- setof(A, agentName(A), [Name|_]).

% priority N  ==>  priority(N)  AND  agent Ag priority N  ==> agent(priority(Ag,N)) 
getAgentPriority(Pri) :- setof(N, bmap_priority(N), Ps) -> last(_,Pri,Ps) ; Pri = 0.
bmap_priority(N):- priority(N), integer(N), N > -1.
% bmap_priority(N) :- agent(priority(_,N)).
% la seconda clausola non serve:
% in sicsplan gli agenti sono anonimi (il nome da usare
% per identificarsi col supervisor e' impostato in settings.pl)


%UNPRETTY FLUENT
% fluent F valued V1,V2   ==> fluent(valued(F,','(V,V)))
fluent(F,MinF,MaxF) :- fluent(valued(F,','(MinF,MaxF))).
% fluent F valued Dom   ==> fluent(valued(F,Dom))
fluent(F,Dom) :- fluent(valued(F,Dom)), \+functor(Dom,',',2).

fluentName(F) :- fluent(F,_MinF,_MaxF); fluent(F,_Dom).

%UNPRETTY ACTION
% action Act  ==>  orig_action(Act)
% (rimpiazza il predicato action/1, per evitare multifile-def)
orig_action(A) :- action(A).
orig_action(A) :- on_conflict(action(A),_).
orig_action(A) :- on_failure(action(A),_).


% action Act on_conflict(provided(retry_after(T), Cs))
% action Act on_conflict(retry_after(T))
% action Act on_conflict(provided(forego, Cs))
% action Act on_conflict(forego)
% action Act on_conflict(arbitration)
% E
% action Act on_failure(if(retry_after(T), Cs))
% action Act on_failure(retry_after(T))
% action Act on_failure(if(replan, add_goal(Cs, Gs)))
% action Act on_failure(if(replan, Cs))
% action Act on_failure(replan)
% action Act on_failure(add_goal(replan, Gs))
% action Act on_failure(fail)
% action Act on_failure(if(fail, Cs))

%ESEMPIO 
%on_conflict(action(act1),arbitration).
%on_conflict(action(act2),arbitration).
%on_conflict(action(act3),arbitration).
%ESEMPIO 
%on_conflict(action(act1),on_conflict(provided(retry_after(1),[f eq 2]), arbitration)).
%on_conflict(action(act2),on_conflict(retry_after(3), arbitration)).
%on_conflict(action(act3),on_conflict(provided(retry_after(3),[g eq 234]), arbitration )).
%ESEMPIO 
%on_conflict(action(act1),provided(retry_after(1),[f eq 2])).
%on_conflict(action(act2),retry_after(3)).
%on_conflict(action(act3),provided(retry_after(3),[g eq 234])).

%ASSUMIAMO UNA SOLA DEF PER OGNI AZIONE ActionName :
%(altrimenti si toglie il CUT e si esegue findall+appends....
on_X_Policies(ActionName, OnCPs,OnFPs) :-
	( ( on_failure(action(ActionName), OnPolicies), onf_split(OnPolicies,OnCPs,OnFPs) )
	; ( on_conflict(action(ActionName), OnPolicies), onc_split(OnPolicies,OnCPs,OnFPs) ) ),
	!.
on_X_Policies(_ActionName, [],[]).

onc_split(on_failure(P1,Ps), [P1aux|CPs], FPs) :- !,
	convertPolicy(P1,conflict,P1aux),
	onf_split(Ps, CPs, FPs).
onc_split(on_conflict(P1,Ps), [P1aux|CPs], FPs) :- !,
	convertPolicy(P1,conflict,P1aux),
	onc_split(Ps, CPs, FPs).
onc_split(P1, [P1aux], []) :- !,
	convertPolicy(P1,conflict,P1aux).

onf_split(on_failure(P1,Ps), CPs, [P1aux|FPs]) :- !,
	convertPolicy(P1,failure,P1aux),
	onf_split(Ps, CPs, FPs).
onf_split(on_conflict(P1,Ps), CPs, [P1aux|FPs]) :- !,
	convertPolicy(P1,failure,P1aux),
	onc_split(Ps, CPs, FPs).
onf_split(P1, [], [P1aux]) :- !,
	convertPolicy(P1,failure,P1aux).


convertPolicy(if(retry_after(T),Cs), failure, if(Clist,retry_after(T))) :-
	integer(T), T>0,
	(commatolist(Cs,T2) -> (Clist=T2) ; (Clist=Cs)).
convertPolicy(retry_after(T), failure, if([1 eq 1],retry_after(T))) :-
	integer(T), T>0.
convertPolicy(if(replan,add_goal(Cs,Gs)), failure, if(Clist,replan(Glist))) :-
	(commatolist(Gs,TG) -> (Glist=TG) ; (Glist=Gs)),
	(commatolist(Cs,TC) -> (Clist=TC) ; (Clist=Cs)).
convertPolicy(if(replan,Cs), failure, if(Clist,replan([1 eq 1]))) :-
	(commatolist(Cs,TC) -> (Clist=TC) ; (Clist=Cs)).
convertPolicy(add_goal(replan,Gs), failure, if([1 eq 1],replan(Glist))) :-
	(commatolist(Gs,TG) -> (Glist=TG) ; (Glist=Gs)).
convertPolicy(if(fail,Cs), failure, if(Clist,replan([1 eq 0]))) :-
	(commatolist(Cs,TC) -> (Clist=TC) ; (Clist=Cs)).
convertPolicy(replan, failure, if([1 eq 1],replan([1 eq 1]))).
convertPolicy(fail, failure, if([1 eq 1], replan([1 eq 0]))).

convertPolicy(provided(retry_after(T),Cs), conflict, retry_after(T,Clist)) :-
	integer(T), T>0,
	(commatolist(Cs,TC) -> (Clist=TC) ; (Clist=Cs)).
convertPolicy(retry_after(T), conflict, retry_after(T)) :-
	integer(T), T>0.
convertPolicy(provided(forego,Cs), conflict, forego(Clist)) :-
	(commatolist(Cs,TC) -> (Clist=TC) ; (Clist=Cs)).
convertPolicy(forego, conflict, forego).
convertPolicy(arbitration, conflict, arbitration).




%UNPRETTY EXECUTABILITY
% executable Act if C,...,C
executable(Act,C) :- unpretty_executable(Act,C).
%executable(if(Act, ','(C,','(C,...))))
%executable(if(Act, [C,C,...]))
unpretty_executable(Act,C):-
    executable(if(Act, Cs)),
    (commatolist(Cs,T2) -> (T2=C) ; (Cs=C)).



%%% STATIC NON CONTEMPLATE PER ORA:
%%%	%UNPRETTY CAUSED RealL=>Constr
%%%	orig_caused(RealL,Constr) :-
%%%	    caused(List,Constr),
%%%	    ( commatolist(List,LL) ->  (LL=RealL) ; (List=RealL) ).


%UNPRETTY CAUSES  RealL=>Constr
%causes Effect if Precondition   ==>  causes(Effect,Precondition)
%causes( if(C, ','(C,','(C,...))) )
%causes( if(C, [C,C,...]) )
causes(Constr,RealL) :-
    causes(if(Constr,List)),
    ( commatolist(List,LL) ->  (LL=RealL) ; (List=RealL) ).
%% BMAP vs BMV
%per rattoppare le differenze dtra le sintassi di BMAP e BMV
%Action causes Effect if Precondition   ==>  causes(Action, if(Effect,Precondition))
causes(ActionName, Constr, RealL) :-
    causes(ActionName, if(Constr,List)),
    ( commatolist(List,LL) ->  (LL=RealL) ; (List=RealL) ).



%UNPRETTY INITIALLY  (works for lists and comma-separated lists of constraints)
% initially  C,...,C    ==>   orig_initially(C)
orig_initially(CoCL) :-
    initially(C),
    (is_list(C) -> member(CoCL,C) ;
     commatolist(C,LL) -> member(CoCL,LL) ;
     (CoCL = C)).

%UNPRETTY GOAL  (works for lists and comma-separated lists of constraints)
% goal  C,...,C    ==>   orig_goal(C)
orig_goal(CoCL) :-
    goal(C),
    (is_list(C) -> member(CoCL,C) ;
     commatolist(C,LL) -> member(CoCL,LL) ;
     (CoCL = C)).


%UNPRETTY ALWAYS  (works for lists and comma-separated lists of constraints)
orig_always(CoCL) :-
    always(C),
    (is_list(C) -> member(CoCL,C) ;
     commatolist(C,LL) -> member(CoCL,LL) ;
     (CoCL = C)).

%UNPRETTY TIME_CONSTRAINT  (works for lists and comma-separated lists of constraints)
orig_time_constraint(CoCL) :-
    time_constraint(C),
    (is_list(C) -> member(CoCL,C) ;
     commatolist(C,LL) -> member(CoCL,LL) ;
     (CoCL = C)).



%UNPRETTY COSTS (=action_cost)
action_cost(Ag,A,CA) :- unpretty_action_cost(Ag,A,CA).
%costs(by(X, ','(A,','(A,...))), E)
%costs(by(X, [A,A,...]), E)
unpretty_action_cost(Agents,Act,E):-
    costs(by(Act, As), E),
    (commatolist(As,T1) -> (T1=Agents) ; (As=Agents)).




%ADDED march 2011 by Andy (copiato da bmap.pl)
%   (TODO da implementare nei conflictSolve*)
%UNPRETTY help  (laws stating when this agent(s)/planner has to help
%                an (external, of another planner) one, among AgentList);
%                default: do not help
% two possible forms: 1)   help AgentList if CondList
%                     2)   help AgentList
%   AgentList might be:  "[a,b,c,...]"  or  "a,b,c,..."  or  "all"
%   CondList  might be:  "[PC1,PC2,PC3,...]"  or  "PC1,PC2,PC3,..."
%   dove PCi sono constraint valutabili, ovvero, per es, senza riferimenti futuri o assoluti
%
% thisAgentHelps_DNF(+OtherAgent, -DNF) instantiate DNF is a list of lists on constraints
% if DNF then this planner/agent can help OtherAgent.
thisAgentHelps_DNF(OtherAgent, DNF) :-
    ground(OtherAgent),
    findall(Conj, one_help_conjunct(OtherAgent,Conj), DNF), !.  %fails if no help-laws exist

one_help_conjunct(OtherAgent, [0 eq 0]) :-  %if success: a law of type 2) exists for OtherAgent
    help(As), As \= if(_,_),
    (
      As == all ;
      ((commatolist(As,T1) -> (T1=AgentLs) ; (As=AgentLs)) , memberchk(OtherAgent,AgentLs))
    ), !. % stop gathering other subsumed laws
one_help_conjunct(OtherAgent, Conj) :- % collect cases of type 1) for OtherAgent
    help(if(As,Cs)),
    (
      As == all ;
      ((commatolist(As,T1) -> (T1=AgentLs) ; (As=AgentLs)) , memberchk(OtherAgent,AgentLs))
    ),
    (commatolist(Cs,T2) -> (T2=Conj) ; (Cs=Conj)).

%ADDED march 2011 by Andy
%acceptableRequest(+RequiringAgentName,+RequiredCond)
%true if this agent can accept the help-request form RequiringAgentName,
%which is requiring to ensure truth of RequiredCond.
acceptableRequest(RequiringAgentName,RequiredCond) :- 
	thisAgentHelps_DNF(RequiringAgentName,PreconditionDNF),
	verifyPrecondition(PreconditionDNF),
	ableToEnsure(RequiredCond).

%TODO ora sempre verificata (andrebbe ricicciato il codice usato in
%conflictResolution* per verificare la soddisfacibilita' di una condizione)
verifyPrecondition(PreconditionDNF) :-
	format("~n****verifyPrecondition-ASK USER: Is this DNF satisfiable? ~w~n Type Y/N?",[PreconditionDNF]),
	read(Ans),!,nl,(Ans==y ; Ans=='Y').

%TODO ora sempre verificata (andrebbe ricicciato il codice usato in
%conflictResolution* per verificare la soddisfacibilita' di una condizione)
ableToEnsure(RequiredCond) :-
	format("~n****ableToEnsure-ASK USER: Can this agent ensure that ~w holds? ~w~n Type Y/N?",[RequiredCond]),
	read(Ans),!,nl,(Ans==y ; Ans=='Y').



%AUX
%commatolist(CommaSep,List)
commatolist(','(X,','(Y,Z)),[X|L]) :- commatolist(','(Y,Z),L).
commatolist(','(X,Y),[X,Y]) :- \+functor(Y,',',2).
commatolist(NotList,[NotList]) :- \+is_list(NotList).




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
meta_planner_multi(StartLen, MaxLen, ActionDeltaLists, DeltasTrajectory, StateSequence) :-
   meta_planner(StartLen,MaxLen,Actions,StateSequence),
   %mkList(Actions,ActionLists),
   %write('STEP  XX:'),read(_),
   mkDTrajectory(StateSequence, Actions, ActionDeltaLists, DeltasTrajectory).

mkList([],[]).
mkList([Act|Acts],[[Act]|Rest]) :-
   mkList(Acts,Rest).

mkDTrajectory([_Last], [], [], []) :-!.
mkDTrajectory([_Last], [X|_], [], []) :- !,
        format("UNEXPECTED extra action ~w in plan! Skipping.~n",[X]).
mkDTrajectory([From,To|States], [MultiAct|MultiActRest], [ActionDeltaList|ActionDeltaLists], [Delta|DeltaSequence]) :-
        getStateChanges(From,To,Delta),
	mkmicrodeltas(MultiAct,Delta,ActionDeltaList),
        mkDTrajectory([To|States], MultiActRest, ActionDeltaLists, DeltaSequence).
getStateChanges([],[],[]).
getStateChanges([[F,V1]|FVs1],[[F,V2]|FVs2],Delta) :-
        V1 == V2,!,
        getStateChanges(FVs1,FVs2,Delta).
getStateChanges([[F,V1]|FVs1],[[F,V2]|FVs2],[[F,V2]|Delta]) :-
        V1 \== V2,!,
        getStateChanges(FVs1,FVs2,Delta).
getStateChanges([[F1,_]|FVs1],[[F2,_]|FVs2],Delta) :-
        F1 \== F2,
        format("UNEXPECTED fluent name while detecting changes in a state transition. Missing fluent? Found ~w and ~w. Skipping one.~n", [F1,F2]),
        ((F1 @< F2) -> getStateChanges(FVs1,[[F2,_]|FVs2],Delta) ; getStateChanges([[F1,_]|FVs1],FVs2,Delta)).

mkmicrodeltas(MultiAct,Delta,ActionDeltaList) :-
	format("~n*********~nSTART EVAL MICRODELTAs DELTA=~w~n   MULTIACT=~w~n", [Delta,MultiAct]),
	mkmicrodeltas_aux(MultiAct,Delta,ActionDeltaList,Delta,_Common),
	format("~n*********~nMICRODELTAs=~w~n", [ActionDeltaList]).

mkmicrodeltas_aux([],_Delta,[],Common,Common). %(1)
mkmicrodeltas_aux([OneAct|OtherActs],Delta,[OneAct-MicroDComm|Rest], Remainder, Common) :-
	mkonemicrodelta(OneAct,Delta,MicroD),
	findall(X, (member(X,Remainder),nonmember(X,MicroD)), NewRemainder),
	append(MicroD,Common,MicroDComm), %qui Common e' ancora _, verra' istanziata da (1) per tutte le liste
%	append(Rest,Common,NewRemainder),
        mkmicrodeltas_aux(OtherActs,Delta,Rest, NewRemainder, Common).

mkonemicrodelta(act([single],OneAct),Delta,MicroD) :-
	findall([F,V], (fluentName(F),causes(OneAct,FC,_LC),zero_subterm(F,FC),member([F,V],Delta)), MicroD).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

meta_planner(Start,Max,[],[[]]) :-
   Start > Max, !, 
   format("~n~n******~n******WARNING: meta_planner(sicsplan): wrong length bounds !!!! ~w~n******~n~n", [' ']).
meta_planner(Start,Max,OUTActionList,OUTStateList) :-
   %write('STEP  II:'),read(_),
   statistics(runtime,[Ts,_]),
   sicsplan_up(Start,Max,OUTActionList,OUTStateList),
   statistics(runtime,[Te,_]),
   T is (Te-Ts)/1000,
   format("***Total time: ~2f s ~n",[T]).

sicsplan_up(NSteps,Max,ActionList,StateList) :-
  NSteps =< Max,
  format("~n*********~nSearching for a plan with length ~w~n*********~n", [NSteps]),
  sicsplan(NSteps,ActionList,StateList),
  !,
  format("~n*********~nPlan of ~w steps found!!~n*********~n", [NSteps]).
sicsplan_up(NSteps,Max,ActionList,StateList) :-
  NSteps < Max,
  NSteps1 is NSteps +1,
  sicsplan_up(NSteps1,Max,ActionList,StateList).
sicsplan_up(NSteps,Max,[],[[]]) :-
  NSteps > Max,!,
  format("~n******~n******Cannot find a plan with at most ~w steps~n******~n", [Max]).




meta_planner_multi(StartLen, MaxLen, ActionDeltaLists, DeltasTrajectory, StateSequence) :-
   meta_planner(StartLen,MaxLen,Actions,StateSequence),
   mkDTrajectory(StateSequence, Actions, ActionDeltaLists, DeltasTrajectory).


meta_plannerPrefixed_multi(StartLen,MaxLen,RevPlanDoneMulti,PlanPrefixTodo,ThisStepNum,NewRevPlanDone,ActionDeltaLists,DeltasTrajectory) :-
	%mkList(RevPlanDoneMulti, RevPlanDone),
	deDTrajectory(RevPlanDoneMulti,RevPlanDone),
	%mkList(PlanPrefixTodoMulti, PlanPrefixTodo), %da controllare in che forma arriva dal supervisor/agreement
	meta_plannerPrefixed(StartLen,MaxLen,RevPlanDone,PlanPrefixTodo,ThisStepNum,NewRevPlanDone,ActionList,StateList),
        mkDTrajectory(StateList, ActionList, ActionDeltaLists, DeltasTrajectory).
     %mkList(NewRevPlanDoneMulti, NewRevPlanDone),
     %mkList(ActionListMulti, ActionList).


deDTrajectory([],[]).
deDTrajectory([Trans|Transs],[Distil|Distils]) :-
	deDTrajectory_aux(Trans,Distil),
	deDTrajectory(Transs,Distils).

deDTrajectory_aux([],[]).
deDTrajectory_aux([Act-_|Acts],[Act|Dist]) :- !,
	deDTrajectory_aux(Acts,Dist).
deDTrajectory_aux([Act|Acts],[Act|Dist]) :- !,
	deDTrajectory_aux(Acts,Dist).



meta_plannerPrefixed(Start,Max,RevPlanDone, _PlanPrefixTodo, _ThisStepNum, RevPlanDone, [], [[]]) :-
   Start > Max, !,
   format("~n~n******~n******WARNING: meta_plannerPrefixed(sicsplan): wrong length bounds !!!! ~w~n******~n~n", [' ']).
meta_plannerPrefixed(Start,Max,RevPlanDone, PlanPrefixTodo, ThisStepNum, NewRevPlanDone, ActionList,StateList) :-
   statistics(runtime,[Ts,_]),
   sicsplanPrefixed_up(Start, Max, RevPlanDone, PlanPrefixTodo, ThisStepNum, NewRevPlanDone, ActionList,StateList),
   statistics(runtime,[Te,_]),
   T is (Te-Ts)/1000,
   format("***Total time: ~2f s ~n",[T]).

sicsplanPrefixed_up(NSteps, Max, RevPlanDone, PlanPrefixTodo, ThisStepNum, NewRevPlanDone, ActionList,StateList) :-
  NSteps =< Max,
  format("~n*********~nSearching for a plan with length ~w~n*********~n", [NSteps]),
  sicsplanPrefixed(NSteps, RevPlanDone, PlanPrefixTodo, ThisStepNum, NewRevPlanDone, ActionList, StateList),
  !,
  format("~n*********~nPlan of ~w steps found!!~n*********~n", [NSteps]).
sicsplanPrefixed_up(NSteps, Max, RevPlanDone, PlanPrefixTodo, ThisStepNum, NewRevPlanDone, ActionList,StateList) :-
  NSteps < Max,
  NSteps1 is NSteps +1,
  sicsplanPrefixed_up(NSteps1, Max, RevPlanDone, PlanPrefixTodo, ThisStepNum, NewRevPlanDone, ActionList,StateList).
sicsplanPrefixed_up(NSteps, Max, RevPlanDone, _PlanPrefixTodo, ThisStepNum, RevPlanDone, NOOPs,EMPTYs) :-
  NSteps >= Max,!,
  Rest is Max-ThisStepNum,
  %format("~n~n******~n***********************************************PRIMA ~w --- ~w --- ~w~n******~n~n", [NSteps,Max,ThisStepNum]),
  format("~n******~n******Cannot find a plan with at most ~w steps~n******~n", [Max]),
  mkNnoops(Rest,NOOPs,EMPTYs).


% MODIF by andy, March 2011
%%%%%%%%
% in caso di fallimento del planner (non e' posibile trovare un piano),
% tramite mkNnoops  si genera la coda del piano eseguendo solo NO-OP.
% Questo piano fittizio in cui l'agente del planner esegue sempre NO-OP,
% non raggiunge il goal, ma e' meglio operare cosi' piuttosto che abbandonare,
% perche' nei prossimi passi lo stato potrebbe cambiare per effetto di altri
% planner/agenti e in una successiva fase di ri-pianificazione
% questo planner potrebbe avere un'altra possibilita' di trovare un piano....
mkNnoops(0,[],[[]]).
mkNnoops(1,['NO-OP'],[[],[]]).
mkNnoops(N,['NO-OP'|NOOPs],[[]|EMPTYs]) :- N>1, N1 is N-1,
    mkNnoops(N1,NOOPs,EMPTYs).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%TODO
% questo planner per BMV non accetta la sintassi di BMAP completa
% Potremmo estendere con il trattamento di caused, globalconstraint 
% (questi ultimi diventano "vincoli ambiantali" (Perugia nov 2010)
% 
%%%% %%%% %%%% %%%% %%%% %%%% %%%% %%%% %%%% %%%% %%%% %%%% %%%% %%%%

%TENTIVO DI RISTRUTTURARE: accordare con gli altri client/agenti e il supervisor....
% Il piano deve essere prodotto come: initial + lista di actions + lista di stati
% e non solo stampato in output....
sicsplan(Steps,ActionList,StateList) :-
    M is Steps + 1,
    retractall(smart_cluster(_,_)),
    main(M,Actionsocc,States),!,
    mkActionSequence(Actionsocc,ActionList),
    mkStateSequence(States,StateList).
sicsplan(N,_,_) :-
    format("~n******~nSearch for plan long ~w failed!~n******~n", [N]), !,fail.


% versione con un prefisso di azioni predeterminato
%sicsplanPrefixed(MaxSteps, RevPlanDone, PlanPrefixTodo, ThisStepNum, NewRevPlanDone, NextActionList, NextDeltas) :-
sicsplanPrefixed(Steps, RevPlanDone, PlanPrefixTodo, ThisStepNum, RevPlanDone, ActionList, StateList) :-
	format("sicsplanPrefixed: reverse prefix done: ~w~n",[RevPlanDone]),
	format("sicsplanPrefixed:         prefix todo: ~w~n",[PlanPrefixTodo]),
    M is Steps + 1,
    retractall(smart_cluster(_,_)),
    mainPrefixed(M, RevPlanDone, PlanPrefixTodo, Actionsocc,States),!,
    length(L,ThisStepNum), append(L,ActionsoccSuff,Actionsocc),
    mkActionSequence(ActionsoccSuff,ActionList),!,
    length(LL,ThisStepNum), append(LL,StatesSuff,States),
    mkStateSequence(StatesSuff,StateList),
	format("sicsplanPrefixed: NEW future plan todo: ~w~n",[ActionList]),
    !.
sicsplanPrefixed(N,_,_,_,_,_,_) :-
    format("~n******~nSearch for plan long ~w failed!~n******~n", [N]), !,fail.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

mkActionSequence([],[]).
mkActionSequence([Trans|Transs],[TransAgsName|TransAgsNames]) :-
    format("~n******~nfind_actions in mkActionSequence on ~w****", [Trans]),
        extract_actions(Trans,TransAgsName),
    format("******FATTA: ~w~n", [TransAgsName]),
        mkActionSequence(Transs,TransAgsNames).

extract_actions([],[]).
extract_actions([action(Name,Value)|Rest], [act([single],Name) |S]) :-
    Value > 0,!,
    extract_actions(Rest,S).
extract_actions([_|Rest], S) :-
    extract_actions(Rest,S).



mkStateSequence([],[]).
mkStateSequence([InState|InStates],[OutSortState|OutSortStates]) :-
	mkOutState(InState,OutState),
	sort(OutState, OutSortState),
	mkStateSequence(InStates,OutSortStates).
mkOutState([],[]).
mkOutState([fluent(Name,Value)|InRest], OutRest) :-
	fd_var(Value),!,
	format("UNEXPECTED undeterminated fluent value for fluent: ~w.  Skipping it.~n", [Name]),
	mkOutState(InRest,OutRest).
mkOutState([fluent(Name,Value)|InRest], [[Name,Value]|OutRest]) :-
	mkOutState(InRest,OutRest).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%main(NumStates) :-
%    main(NumStates,_,_),!.
%main(_) :-
%    format("No solutions.~w~n",[' ']).


main(NumStates, Actionsocc, States) :-
    build_states(NumStates,States,Nf),
    make_action_occs(NumStates,Actionsocc,PlanCost,Na),
    format("There are ~q fluents per state and ~q possible actions~n",[Nf,Na]),
    mainCore(NumStates, Actionsocc, States, PlanCost).
    

mainPrefixed(NumStates, RevPlanDone, PlanPrefixTodo, Actionsocc, States) :-
    build_states(NumStates,States,Nf),
    make_action_occs(NumStates,Actionsocc,PlanCost,Na),
    reverse(RevPlanDone,Done), append(Done, PlanPrefixTodo, Prefix),
    fixPrefix_actions(Prefix,Actionsocc),
    % length(PlanPrefix,P),
    % disableSubsequentNOOPs(P, Actionsocc),
    format("There are ~q fluents per state and ~q possible actions~n",[Nf,Na]),!,
    mainCore(NumStates, Actionsocc, States, PlanCost).


fixPrefix_actions([],_).
fixPrefix_actions([Actions|As], [OneTransFlags|OtherActionsocc]) :-
	fixOneActionList(Actions,OneTransFlags),!,
	fixPrefix_actions(As,OtherActionsocc).

fixOneActionList([],_OneTrans).
fixOneActionList([act([_Agent],Aname)|Acts],OneTrans) :-
	fixOneAction(Aname,OneTrans),
        fixOneActionList(Acts,OneTrans).

fixOneAction(Aname,ActOccs) :-
	((member(action(Aname,VA), ActOccs)) -> (VA #= 1) ; (format("WARNING, trying to force the unknown action ~w~n",[Aname]))),!.


mainCore(_NumStates, Actionsocc, States, PlanCost) :-
%%%%%%%%%%%%%%%%     ADD CONSTRAINTS     %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  format("~n~n~nBEFORE set_initial ON STATES:~n~w~n",[States]),
    (set_initial(States); %%% Help in A.T. DEBUG
      write('Inconsistent initial state'),nl,!,fail),
    write('Initial State:'),nl, States=[S|_],write_state(S),nl,
%  format("~n~n~nAFTER set_initial ON STATES:~n~w~n",[States]),
    (set_goal(States); %%% Help in A.T.  DEBUG
      write('Inconsistent final state'),nl,!,fail),
%  format("~n~n~nAFTER set_goal ON STATES:~n~w~n",[States]),
    (set_intermediate_states(States); %%% Help in A.T.  DEBUG
      write('Inconsistent intermediate state'),nl,!,fail),
%  format("~n~n~nAFTER set_intermediate_states ON STATES:~n~w~n",[States]),
    (set_cross_constraints(States); %%% Help in A.T.  DEBUG
      write('Inconsistent cross constraint'),nl,!,fail),
%  format("~n~n~nAFTER set_cross_constraints ON STATES:~n~w~n",[States]),
    (set_action_bounds(Actionsocc);
      write('Check Action Bounds '),nl,!,fail),
    write('Final State:'),nl, append(_,[Last],States),write_state(Last),
    !,
%  format("~n~n~nBEFORE STARTING set_transitions ON STATES:~n~w~n",[States]),
    set_transitions(Actionsocc,States),
    write('****transitions   set****'),nl,
    set_executability(Actionsocc,States),
    write('****executability set****'),nl,
    %   BMAP vs BMV  cost_constrain disabilitati...
    %   set_cost_constraints(States,PlanCost,GC),
    !,
    %write('STEP  II:'),read(_),

%%%%%%%%%%%%%%    LABELING     %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% REMARK Observe that only action variables are labelled.
%%% Fluent variables obtain their values by propagation
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    get_all_actions(Actionsocc, AllActions),
    length(AllActions,L),write('Still to label '),write(L),write(' variables'),nl,
    statistics(runtime,[_,PT]),
    write('Constraints added in ms: '),write(PT),nl,!,

    % standard labeling([ff,min],AllActions),
    my_labeling(Actionsocc,States),

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    statistics(runtime,[_,Time]),
    dump_result(Actionsocc,States),
    write('Total Plan Cost = '),write(PlanCost),nl,
    %   BMAP vs BMV  cost_constrain disabilitati...
    % write('Goal Cost = '),write(GC),nl,
    write('Solution found in Time: '),write(Time),nl.




%%% my_labeling implements a leftmost strategy.
%%% After a new state is reached, it is checked its
%%% difference wrt the prevoius ones

my_labeling(Actionsocc,States) :-
     my_labeling(Actionsocc,States,1).
my_labeling([],_Actionsocc,_States) :- !.%,
    %
%    format("~nEND OF STATES ~w~n~n~n",[' ']).%,
    %
my_labeling([[action('NO-OP',NOOPvar)|CurrAct] | Actions], States,I) :-
    %
%    format("~n~n~nBEFORE STARTING LABELING FOR TRANSACTION ~w ON STATES:~n~w~n",[I,States]),
    %
    (NOOPvar = 0 ; NOOPvar = 1),
    my_labeling_aux(CurrAct),
    ((NOOPvar = 0) -> no_loop(States,I) ; true),
    I1 is I + 1,
    my_labeling( Actions , States,I1).
my_labeling_aux([]).
my_labeling_aux([action(_,A)|R]) :-
    indomain(A),
    my_labeling_aux(R).

%%% no_loop : disabled to allow NO-OP
%TODO:
% riscrivere no_loop in modo che sia applicato solo al suffisso "libero" del plan
% (escludendo cioe' sia il prefisso nel passato che la parte delle azioni future
%  "promised" (queste potrebbero non modificare lo stato locale))

no_loop(_,_).

% no_loop(States,A) :-
%     state_select(A,States,StateA),
%     no_loop(A,States,StateA).
% no_loop(0,_,_):-!.
% no_loop(B,States,StateA) :-
%     B1 is B - 1,
%     state_select(B1,States,StateB),
%     StateA \== StateB, %%% ev. \==
%     no_loop(B1,States,StateA).






setFluentsInState([],_).
setFluentsInState([[F,Val]|Rest],StateNum) :-
	((fluent(F,_) ; fluent(F,_,_)) -> 
		(
			format("SETTING FLUENT VALUE:  ~w  by   (~w) ~n", [[F,Val], F eq Val]),
			asserta(holds(F eq Val , StateNum)),
			asserta(exogenous(F, StateNum))
		)
		; 
		(
			format("SKIP FLUENT ASSIGN:  ~w  ~n", [[F,Val]]),
			true
		)
		),
	setFluentsInState(Rest,StateNum).




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% The domain of Each Fluent Variable is set here
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

build_states(N,States,Nf) :-
    findall([F,int(MinF,MaxF)],fluent(F,MinF,MaxF),L1),
    findall([F,set(Dom)],fluent(F,Dom),L2),
    append(L1,L2,Lf),
    length(Lf,Nf),
    make_states(N,Lf,States).
make_states(0,_,[]):-!.
make_states(N,List,[S|STATES]) :-
    N1 is N-1, make_states(N1,List,STATES),
    make_one_state(List,S).
make_one_state([],[]).
make_one_state([[F,int(MinF,MaxF)]|Fluents],[fluent(F,VarF)|VarFluents]) :-
    make_one_state(Fluents,VarFluents),
    VarF in MinF..MaxF. %%% interval case
make_one_state([[F,set(Dom)]|Fluents],[fluent(F,VarF)|VarFluents]) :-
    make_one_state(Fluents,VarFluents),
    VarF in Dom.  %%% set case

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% The domain of each action (occurrence) variable is set to {0,1}
%%% In each state transition exactly one action occur (sum constraint)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

make_action_occs(N,Actionsocc,PlanCost,Na) :-
    setof(A,orig_action(A),La1),
    La = ['NO-OP'|La1],  %ADDED for replanning
    length(La,Na),
    make_action_occurrences(N,La,Actionsocc,PlanCost).
make_action_occurrences(1,_,[],0):-!.
make_action_occurrences(N,List,[Act|ActionsOcc],Cost) :-
    N1 is N-1,
    make_action_occurrences(N1,List,ActionsOcc,Cost1),
    make_one_action_occurrences(List,Act,Cost2),
    get_action_list(Act,AList),
    fd_only_one(AList), %% fd_one_or_two(AList),
    Cost #= Cost1 + Cost2.
make_one_action_occurrences([],[],0).
make_one_action_occurrences([A|Actions],[ action(A,OccA)|OccActions],Cost) :-
    make_one_action_occurrences(Actions,OccActions,Cost1),
    fd_domain_bool(OccA),
    (action_cost(A,CA),!; CA = 1), %%%Default action cost = 1
     Cost #= OccA*CA + Cost1.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Initial state info are set here.
%%% In the first state static causal rules are applied by "complete_state"
%%% (for the other states they are applied in set transition)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

set_initial([Initial|_]) :-
    findall([FE1,OP,FE2],(orig_initially(C),C=..[OP,FE1,FE2]), InitFC),
    set_this_state(InitFC,Initial),
    complete_state(Initial).

set_goal(States) :-
    findall([FE1,OP,FE2],(orig_goal(C),C=..[OP,FE1,FE2]), GoalFC),
    length(States,N),M is N-1,
    set_state(M,GoalFC,States).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% intermediate constraints are set here
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

set_intermediate_states(States) :-
    findall([FE1,OP,FE2],(orig_always(C),C=..[OP,FE1,FE2]), Always),
    set_always(Always,States),
    findall([FE1,OP,FE2,No],(holds(C,No),C=..[OP,FE1,FE2]), Holds),
    set_fixed(Holds,States).

set_cross_constraints(States) :-
    findall([FE1,OP,FE2],(orig_time_constraint(C),C=..[OP,FE1,FE2]), Cross),
    set_cross_constraints(Cross,States).

%%%%%%% AUXILIARY Predicates
%%% 1) for "cross" constraints

set_cross_constraints([],_).
set_cross_constraints([[FE1, OP, FE2]|Rest],States) :-
     rel_parsing(FE1,Val1,_,States),
     rel_parsing(FE2,Val2,_,States),
     add_constraint(Val1,OP,Val2),
     set_cross_constraints(Rest,States).

%%% 2) for "always" constraints

set_always([],_).
set_always([FC |Always], States) :-
     length(States,N),
     rec_set_always(0,N,FC,States),
     set_always(Always,States).

rec_set_always(M,M,_,_).
rec_set_always(M,N,FC,States) :-
     M < N, M1 is M + 1,
     set_state(M,[FC],States),
     rec_set_always(M1,N,FC,States).

%%% 3) for "holds" constraints

set_fixed([],_).
set_fixed([[FE1,OP,FE2,No]|Rest],States) :-
     state_select(No,States,StateI),
     set_this_state([[FE1,OP,FE2]],StateI),
     set_fixed(Rest,States).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

set_action_bounds(Actionsocc) :-
    findall([ACT,OP,NUM],(action_bound(AC),AC =..[OP,ACT,NUM]), L),
    set_action_bounds(L, Actionsocc).
set_action_bounds([],_).
set_action_bounds([[ACT,OP,NUM]|R],Actionsocc):-
    list_var(Actionsocc,ACT,VARS),
    sum(VARS,OP,NUM),
    set_action_bounds(R,Actionsocc).
list_var([],_,[]).
list_var([AL|R],ACT,[V_ACT|VARS]) :-
    member(action(ACT,V_ACT),AL),
    list_var(R,ACT,VARS).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%% set fluent constraints on state I   %%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

set_state(I,FluentConstraints,States) :-
     state_select(I,States,StateI),
     set_this_state(FluentConstraints,StateI).

set_this_state([],_).
set_this_state([[FE1,OP,FE2]|Rest],State) :-
     parsing(FE1,Val1,State),
     parsing(FE2,Val2,State),
     add_constraint(Val1,OP,Val2),
     set_this_state(Rest,State).

complete_state(State) :-
    complete_state(State,State).
complete_state([],_).
complete_state( [fluent(Fluent,_)| Fluents],InitialState) :-
    set_one_static_fluent(Fluent, InitialState),
    complete_state(Fluents ,InitialState).
set_one_static_fluent(Name, State) :-
     findall([OP,FE1,FE2,L],
         (caused(L,FC),subterm(Name,FC),FC =..[OP,FE1,FE2]), Stat),
     static(Stat,[State],0,_).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% BERLIN-VENEDIG CHANGES - Sept 19 2009
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

set_cost_constraints(States, PlanCost, GOALCOST) :-
     set_goalcost(States, GOALCOST),
     set_plancost(PlanCost),
     set_statecosts(States).

set_goalcost(States, GC) :-
    findall([OP,Num],(cost_constraint(C), C =.. [OP,goal,Num]), GoalCosts),
    length(States,N),M is N-1,
    state_select(M,States,FinalState),
    (state_cost(FE),!; FE = 1), %%% default state cost = 1
    parsing(FE,GC,FinalState),
    set_goalcost_aux(GoalCosts,GC).
set_goalcost_aux([],_).
set_goalcost_aux([[OP,Num]|GoalCosts],GC) :-
      add_constraint(GC,OP,Num),
      set_goalcost_aux(GoalCosts,GC).

set_plancost(PC) :-
    findall([OP,Num],(cost_constraint(C), C =.. [OP,plan,Num]), PlanCosts),
    set_plancost_aux(PlanCosts,PC).
set_plancost_aux([],_).
set_plancost_aux([[OP,Num]|PlanCosts],PC) :-
     add_constraint(PC,OP,Num),
     set_plancost_aux(PlanCosts,PC).


set_statecosts(States) :-
    findall([I,OP,Num],(cost_constraint(C), C =.. [OP,state(I),Num]), StateCosts),
    set_statecost_aux(StateCosts,States).
set_statecost_aux([],_).
set_statecost_aux([[I,OP,Num]|StateCosts],States) :-
    (state_cost(FE),!; FE = 1),
    rel_parsing(FE,Val,I,States),
    add_constraint(Val,OP,Num),
    set_statecost_aux(StateCosts,States).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% constraints on transitions are set here. First each tranisition
%%% is selected. Then every fluent is analyzed and its new value
%%% is constrained using its previous value and the dynamic and static
%%% rules applied in that state transition.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

set_transitions(ActOccs,States) :-
    set_transitions(ActOccs,0,States).

set_transitions([],_,_) :- !.
set_transitions([Occ|Occs], M, States) :-
%  format("~n~n~nIN set_transitions NUMBER ~w :~n~w~n",[M,States]),
    state_select(M, States, StateM),
    set_transition(StateM, M, Occ, States),
    M1 is M + 1,
    set_transitions(Occs, M1, States).

set_transition([],_,_,_).
set_transition([Fluent|RStateM],M,Occ,States) :-
    set_one_fluent(Fluent, M, Occ, States),
    set_transition(RStateM, M, Occ, States).

%%% REMARK. It works with references on previous states
%%% For future references, the list L below must be computed
%%% in a more complex way

set_one_fluent(fluent(FluentName,IV), M,Occ,States) :-
%  format("IN set_one_fluent WORKING ON FLUENT ~w :~n",[fluent(FluentName,IV)]),
    findall([Act,OP,FE1,FE2,L],
    (causes(Act,FC,L),zero_subterm(FluentName,FC),FC =.. [OP,FE1,FE2]),Dyn),
%  format("IN set_one_fluent DYNAMs FOR FLUENT ~w :~n Dyn: ~w ~n",[fluent(FluentName,IV),Dyn]),
    state_select(M,States,StateM), %%% StateM = FromSt
    M1 is M + 1,
    state_select(M1,States,StateM1), %%% StateM1 = ToSt
    member(fluent(FluentName,EV),StateM1),
% format("STEP before dynamic~w",[' ']), read(_),
    dynamic(Dyn,Occ, StateM,DynFormula, M1, States),
    findall([OP,FE1,FE2,L],
%    %% OLD
%     %%  (caused(L,FC),zero_subterm(FluentName,FC), FC =..[OP,FE1,FE2]),
%    %% NEW:
            (( smart_cluster(FluentName,Fluents),!;
               cluster([FluentName], Fluents),
               assert( smart_cluster(FluentName,Fluents) )),
             member(F,Fluents), caused(L,FC),
             zero_subterm_all(F,[L,FC]), FC =..[OP,FE1,FE2]),
        Stat),
    %%% write('###'),write(FluentName),write(Stat),nl,
    sort(Stat,Statord),
    static(Statord, States, M1, StatFormula),
    bool_disj(DynFormula,StatFormula,Formula),
%    bool_disj(DynFormula,Formula),

%rattoppo per l'effetto "esogeno": alcuni fluenti che possono esser modificati da altri agenti
    fd_domain_bool(Exo),
    (exogenous(FluentName,M1) -> (Exo#=1) ; (Exo#=0) ),

%%% inertia
    #\ (Formula #\/ Exo) #=> EV #= IV.


dynamic([],_,_,[],_,_).
dynamic([[Act,OP,FE1,FE2,Prec]|Rest],Occurrence, State,[ Flag |PF1], M, States):-
    member(action(Act,VA),Occurrence),
    N is M - 1, %%% Looks for preconditions in FromSt and back
    get_precondition_vars(N,Prec,States,ListPV),
    length(Prec,NPrec),
    sum(ListPV, SumPrec),
    %%% The effect is in the next state (M = N + 1)
    rel_parsing(FE1,Val1,M,States),
    rel_parsing(FE2,Val2,M,States),
    exp_constraint(Val1,OP,Val2,C),
    (VA  #/\ (SumPrec #= NPrec)) #<=> Flag, %%%%
    Flag #=> C,
    dynamic(Rest,Occurrence, State,PF1, M, States).


static([],_,_,[]).
static([ [OP,FE1,FE2,Cond] | Others], States, M, [Flag |Fo] ) :-
    get_precondition_vars(M,Cond,States,List), % rel
    length(List, NL),
    sum(List, Result),
    rel_parsing(FE1,Val1,M,States),
    rel_parsing(FE2,Val2,M,States),
    exp_constraint(Val1,OP,Val2,C),
    (Result #= NL) #<=> Flag ,
    Flag #=> C,
    static(Others,States,M,Fo).


%%% fixpoint call

cluster(A,B) :-
    collect_fluents_aux(A,C),
    (sort(A,X),sort(C,X), %%% A is equal to C
     !, B=C;
    cluster(C,B)).

collect_fluents_aux([],[]).
collect_fluents_aux([F|Fs],Fluents) :-
    findall(Fnext,( caused(L,C),
                    zero_subterm_all(F,[L,C]),
                    fluent(Fnext,_,_),F \== Fnext,
                    zero_subterm_all(Fnext,[L,C])
                    ), FNEXTS),
    collect_fluents_aux(Fs,Faux),
    union(Faux, [F|FNEXTS], Fluents).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% The executability of each action is then related to the
%%% values of the fluents in the previous state. "formula"
%%% store the disjunction of all executability conditions
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

set_executability(ActionsOcc,States) :-
    findall([Act,C],executable(Act,C),Conds1),
    Conds =[['NO-OP', [0 eq 0]]|Conds1],  %ADDED for replanning
    group_cond(Conds,GroupedConds),
    set_executability(ActionsOcc,0,States,GroupedConds). %%%

set_executability([],_,_,_).
set_executability([ActionsOcc|ARest],M,States,Conds) :-
    set_executability_sub(Conds,ActionsOcc,M,States),
    M1 is M + 1,
    set_executability(ARest,M1, States,Conds).

set_executability_sub([],_,_,_).
set_executability_sub([[Act,C]|CA],ActionsOcc,M,States) :-
    member(action(Act,VA),ActionsOcc),
    preconditions_flags(C, M, States,Flags), % M
    bool_disj(Flags,F),
    VA #=> F,
    set_executability_sub(CA,ActionsOcc,M,States).

preconditions_flags([],_,_,[]).
preconditions_flags([C|R],M,States,[Flag|Flags]) :-
      get_precondition_vars(M,C,States,Cs),
      length(Cs,NCs),
      sum(Cs, SumCs),
      (NCs #= SumCs)  #<=> Flag,
      preconditions_flags(R,M,States,Flags).

get_precondition_vars(_,[],_,[]).
get_precondition_vars(M,[FC|Rest],States,[F|LR]) :-
    FC =.. [OP,FE1,FE2],
    get_precondition_vars(M,Rest,States,LR),
    rel_parsing(FE1,Val1,M,States),
    rel_parsing(FE2,Val2,M,States),
    exp_constraint(Val1,OP,Val2,F).

%%%%%% AUXILIARY predicates

state_select(I,States,StateI) :-
    (nth0(I,States,StateI),!;
    J is I + 1, nth(J,States,StateI)).
%%%%%% compatibility: nth in SICStus 3 / nth0 in SICStus 4

get_all_actions(A,B) :-
   term_variables(A,C),
   reverse(C,B).

get_action_list([],[]).
get_action_list([action(_,V)|Rest],[V|MRest]) :-
    get_action_list(Rest,MRest).

group_cond([],[]).
group_cond([[Action,C]|R],[[Action,[C|Cs]]|S]) :-
     findall(L,(member([Action,L],R)),Cs),
     filter(Action,R,Others),
     group_cond(Others,S).
filter(_,[],[]).
filter(A,[[A,_]|R],S) :-
      !, filter(A,R,S).
filter(A,[C|R],[C|S]) :-
      !, filter(A,R,S).

bool_disj(A,C) :-
  sum(A,B),
  C #<=> B #> 0.

bool_disj(A,B,C) :-
  append(A,B,T),
  bool_disj(T,C).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% SYNTAX and PARSING OF FLUENT EXPRESSIONS:
%%% FE ::= INTEGER | FD_VAR | FE1 * FE2 | FE1 + FE2 | FE1 - FE2 | FE1 / FE2 | FE1 mod FE2
%%%        abs(FE) | rei( FE OP FE) --- reified constraint
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

parsing( Fluent, Val, State ) :-
   rel_parsing(Fluent,Val,0,[State]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% rel_parsing( +FE1,-Val1,+M,+States)

rel_parsing( Num , Num , _, _ ) :-
   integer(Num), !.

rel_parsing( rei(RC), Val, M, States) :-
   RC =.. [OP,E1,E2],
   rel_parsing( E1, Val1, M, States),
   rel_parsing( E2, Val2, M, States),
   exp_constraint(Val1,OP,Val2,Val),
   !.

rel_parsing( abs(FE), Val, M, States) :-
   rel_parsing(FE, Val1, M, States),
   Val #= abs(Val1),
   !.

rel_parsing( FE, Val, M, States ) :-
   FE =.. [OP,FE1,FE2],
   member(OP,[+,-,mod,/,*]),
   rel_parsing( FE1, Val1, M, States),
   rel_parsing( FE2, Val2, M, States),
   ( OP = + -> Val #= Val1 + Val2;
     OP = - -> Val #= Val1 - Val2;
     OP = * -> Val #= Val1 * Val2;
     OP = / -> Val #= Val1 / Val2;
     OP = mod -> Val #= Val1 mod Val2),
   !.

rel_parsing( Fluent^Delta, Val, M, States ) :-
   H is M + Delta,
   length(States,SL),
   in_interval(H,SL,E),
   state_select(E, States, State),
   member(fluent(Fluent,Val),State),
   !.

rel_parsing( Fluent @ Time, Val, _, States ) :-
   state_select(Time,States,State),
   member(fluent(Fluent,Val),State),
   !.

%%% Fluent = Fluent @ M = Fluent^0

rel_parsing( Fluent, Val, M, States ) :-
   state_select(M,States,State),
   member(fluent(Fluent,Val),State).

%%% SYNTAX and BUILDING of a FLUENT CONSTRAINT
%%% FE1 op  FE2, where op in {eq,neq,leq,geq,lt,gt}
%%% %%% %%% %%% %%% %%% %%% %%% %%% %%% %%% %%%

exp_constraint(L,OP,R,C) :-
   (OP == eq  -> C #<=> L #=  R;
    OP == neq -> C #<=> L #\= R;
    OP == geq -> C #<=> L #>= R;
    OP == leq -> C #<=> L #=< R;
    OP == gt  -> C #<=> L #>  R;
    OP == lt  -> C #<=> L #<  R).

add_constraint(L,OP,R) :-
   (OP == eq  -> L #=  R;
    OP == neq -> L #\= R;
    OP == geq -> L #>= R;
    OP == leq -> L #=< R;
    OP == gt  -> L #>  R;
    OP == lt  -> L #<  R).

%%% END MAIN CODE %%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% AUXILIARY PREDICATES for ACTION LANGUAGES
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

neq(A,B) :- A \== B.
diff(A,B) :- A \== B.
diff(A,B,C) :- A \== B, A \== C, C \== B.
interval(A,A,_).
interval(X,A,B) :- A < B, C is A + 1, interval(X,C,B).

in_interval(H,_,0)   :- H <  0, !, write('###warning: reference before beginning of the plan'),nl.
in_interval(H,SL,SL) :- H > SL, !, write('###warning: reference after end of plan'),nl.
in_interval(H,_,H).

%%%%%

subterm(A,B) :-
   A==B,!.
subterm(_,B) :-
   var(B),!,fail.
subterm(A,G) :-
   G =.. [B|_],
   A == B,!.
subterm(A,G) :-
   G =.. [_|Args],
   subterm_all(A,Args).
subterm_all(A,[B|_]) :-
   subterm(A,B),!.
subterm_all(A,[_|R]) :-
   subterm_all(A,R).

%%% it looks for a current fluent

zero_subterm(A,B) :-
   A==B,!.
zero_subterm(_,B) :-
   var(B),!,fail.
zero_subterm(A,G) :-
   G =.. [B|_],
   A == B,!.
zero_subterm(A,A^0) :-
   !.
zero_subterm(A,A^_) :-
   !, fail.
zero_subterm(A,G) :-
   G =.. [_|Args],
   zero_subterm_all(A,Args).
zero_subterm_all(A,[B|_]) :-
   zero_subterm(A,B),!.
zero_subterm_all(A,[_|R]) :-
   zero_subterm_all(A,R).

%%% disjoint append

union([],L,L).
union([A|R],S,T) :-
    member(A,S),!,
    union(R,S,T).
union([A|R],S,[A|T]) :-
    union(R,S,T).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% PRINT
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

dump_result(A,S) :-
    line,
     dump_result(0,A,S),nl,nl,
    line,
    write('Plan in short:'),nl,
    state_plan(A),nl,line.

state_plan([]) :- write('stop.'),nl.
state_plan([A|As]) :-
    find_action(A,Name),
    format("~q -> ",[Name]),
    state_plan(As).


dump_result(N,[],[S]) :-
    write('Time '),write(N),write(': '),
    write_state(S).
dump_result(N,[A|B],[S|Rest]) :-
    write('Time '),write(N),write(': '),
    write_state(S),
    write_action(A),
    N1 is N + 1,
    dump_result(N1,B,Rest).


write_state([]) :- nl.
write_state([fluent(Name,Value)|Rest]) :-
    ( fd_var(Value),!, format("~q: <>  ",[Name]);
      format("~q = ~q\t",[Name,Value]) ),
    write_state(Rest).

write_action(A) :-
    find_action(A,Name),
    format(" ---->>   ~q ",[Name]),nl.

find_action([],unknown).
find_action([action(Name,Value)|_], Name) :-
    Value == 1,!.
find_action([_|Rest],Name) :-
    find_action(Rest,Name).

line :-
    write('*****************************************************************************'),
    nl.

write_list([]).
write_list([A|R]) :- write(A),nl,write_list(R).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%% PORTING GNU => SICStus

%%% Replaced #==> with  #=>
%%% changed labeling -was
%%% fd_labeling(AllActions,[variable_method(ff), value_method(min)])
%%% REPLACED
%%% sum([],0).
%%% sum([A|B],Res) :- sum(B,Res1), Res #= A + Res1.
%%% with

sum(X,Res) :- sum(X,#=,Res).

%%% ADDED

fd_domain_bool(X) :- X in 0..1.
fd_only_one(X) :- sum(X,#=,1).
fd_one_or_two(X) :- sum(X,#=,AA), AA  in 1..2.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%    END CODE    %%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
