:- module(supervisor, [auto_run_supervisor/0,runSupervisor/5,closeSession/0]).

:- use_module('settings.pl').
:- use_module(library(lists)).
:- use_module(m_lindaclient).
:- use_module(library(system)).


% different ways to resolve conflicts (see settings.pl)
:- conflictResolutionStrategy(_N, SupervisorModuleFile, _), use_module(SupervisorModuleFile).




auto_run_supervisor :- 
    %write('STEP ME (type an atom):'),read(_),
    temporaryFileName(FileName),
    see(FileName),
    read(connectionAddress(Address,Port)),
    read(numberOfAgents(NumAgents)),
    read(maxNumberOfSteps(NumSteps)),
    seen,
    runSupervisor(FileName, Address,Port, NumAgents, NumSteps),
    closeSession.
    %halt.
%   !, read(_).


runSupervisor(AddrFilename,Address,Port,NumAgents,NumSteps) :-
   format(" Supervisor: AddrFilename = ~w~n",[AddrFilename]),
   format("             Address      = ~w~n",[Address]),
   format("             Port         = ~w~n",[Port]),
   format("             NumAgents   = ~w~n",[NumAgents]),
   format("             NumSteps     = ~w~n",[NumSteps]),
   %write('STEP ME (type an atom):'),read(_),
   format("Waiting 3 sec... for housekeeping.... ~n",[]),
   sleep(3),
   m_linda_client(Address:Port),
   %testCommunication.
   superviseAgents(NumAgents,NumSteps).

superviseAgents(NumAgents,NumSteps) :-
%format(" Supervisor : TEST 1, digita...~n"), read(_),
    setInitialState(NumAgents,Tckts,Conflicts,InitialState),
    ((Conflicts == []) -> ( performSteps(Tckts,1,NumSteps,InitialState,FinalState),
                endingStuff(Tckts,FinalState) )
               ;  ( true ) ),
    format("Supervisor has done (type 'closeSession/0' after clients ended)~n",[]).
%format(" Supervisor : TEST 2, digita...~n"), read(_),
%closeSession,halt.


setInitialState(NumAgents,Tckts,Conflicts,GlobalInitialState) :-
    format("Supervisor sending START atom...~n",[]),
    m_out(startSettingInitialState),
    %write('STEP ME (type an atom):'),read(_),
    format("Supervisor waiting for agents availability...(sleep(3))...~n",[]),
    %   sleep(3),
    getAllAgentsNames(NumAgents,NamePriority_s),
    format("Supervisor sends tickets to agents...~n",[]),
    sendAllAgentsTickets(NamePriority_s,Tckts),
    format("Supervisor gets initial states from each agent...~n",[]),
    getAllAgentsInitialStates(Tckts, ReordTckts,States),
    format("Supervisor verifies coherence...~n",[]),
    mergeInitialStates(ReordTckts,States,Fluents,GlobalInitialState,Conflicts),!,
%format("XXXXXXX Supervisor detected Conflicts:XXX~wXXX~n", [Conflicts]),
    (Conflicts == [] -> ( format("Supervisor obtained initial state:~w~n", [GlobalInitialState]),
                  format("           fluents:~w~n", [Fluents]),
                      multi_outInitialStateResponse(Tckts,GlobalInitialState),
                      format("Supervisor has set the initial state~n~n",[]) )
             ;  ( format("Supervisor detected inconsistencies in initial state. Conflicts:~w~n", [Conflicts]),
                  multi_outIncoherentInitialStateResponse(Tckts,Conflicts) ) ).



sendAllAgentsTickets([],[]).
sendAllAgentsTickets([Name-Pri|Names],[Tck|Tckts]) :-
    %format("  sending ticket to agent ~w~n",[Name]),
    Tck = ticketToRide(Name,Pri),
    m_out(ticket(Name,Tck)),
    sendAllAgentsTickets(Names,Tckts).


% (DA ELIMINARE???? e utilizzare quello esportato dal planner, ma mediato da plan_executor?)
% lower P is better (i.e. preferred)
getTheAgentPriority(ticketToRide(_N,P),P) :- !.
getTheAgentPriority(_T,0). %default = 0. This should not happen...



getAllAgentsNames(NumAgents,XNames) :-
    getNames(NumAgents, Ns, Ps),
    keys_and_values(NsPs, Ns, Ps),
    sort(NsPs,Names),
    selectLowerPri(Names,XNames).
getNames(0, [], []).
getNames(N, [AgName|Names], [Priority|Priorities]) :-
    m_in(agentReady(AgName,Priority)),
    ((\+ ground(agentReady(AgName,Priority))) -> 
       format("***~nSupervisor: *** WARNING: NON-GROUND AGENT NAME ~w  !!!~n***~n",[agentReady(AgName,Priority)]) ; true ),
    N1 is N-1,
    getNames(N1, Names, Priorities).

selectLowerPri([],[]).
selectLowerPri([Name-Pri],[Name-Pri]) :- !.
selectLowerPri([Name-_Pri1,Name-Pri2|Rest],XNames) :- !,
    selectLowerPri([Name-Pri2|Rest],XNames).
selectLowerPri([NP|Rest],[NP|XNames]) :- !,
    selectLowerPri(Rest,XNames).


expectedTermsBis([],[],_Template).
expectedTermsBis([Tck|Tckts],[Term|R],Template) :-
    copy_term(Template,Term),
    Term =.. [_,Tck|_],
    expectedTermsBis(Tckts,R,Template).


getAllAgentsInitialStates(Tckts, ReorderedTckts,States) :-
    expectedTermsBis(Tckts, List, initialState(_Tck,_InitialPairs)),
%   format("        TEST PRESENZA    ~n"),
%   m_rd(initialState(B,C)),
%   format("        PRESENTE       =~w~n", [initialState(B,C)]),
    length(Tckts, N),
    getInitialStates(N, List, Tckts, ReorderedTckts, States).

%one for each Tck, the order does not care
getInitialStates(0, _List, [], [], []).
getInitialStates(N, List,  Tckts, [ATck|ATckts], [AState|States]) :-
%   format("getInitialStates: chiamo m_in(List, initialState(ATck,AState))  ~n"),
%   format("        List           =~w~n",[List]),
    m_in(List, initialState(ATck,AState)),
%   format("        RICEVUTO       =~w~n", [initialState(ATck,AState)]),
    selectchk(ATck, Tckts, TcktsRest),
    N1 is N-1,
    getInitialStates(N1,List,TcktsRest,ATckts,States).



mergeInitialStates(Tckts,States,Fluents,GlobalInitialState,ConflictsFAs) :-
    findTheConflicts(Tckts,States,Fluents,ConflictsFAs),
    ((ConflictsFAs == []) -> ( append(States,GSs),
                   sort(GSs,GlobalInitialState))
                  ;  ( GlobalInitialState = []) ).


%findTheConflicts(+Tckts,+ListsOfPairs,-Fluents,-Conflicts)
findTheConflicts(Tckts,ListsOfPairs,Fluents,Conflicts) :-
    mkTriples(Tckts,ListsOfPairs,FAVs),
    sort(FAVs, Triples),
    keys_and_values(Triples, Fls, _), sort(Fls, Fluents),
    findConflicts(Fluents, Triples, Conflicts).



%Res is a list of  F-ConflSubset where F is a Fluent and ConflSubset is the list of Agents disagreeing on F
findConflicts([], _, []).
findConflicts([F|Fs], Triples, Res) :-
    findall(A-V, member(F-(A-V),Triples), AVsForF),
    (isConflicting(AVsForF, ConflSubset) -> (Res=[F-ConflSubset|Res1])
                         ;  (Res=Res1) ),
    findConflicts(Fs, Triples, Res1).



isConflicting(AVsForF, ConflSubset) :-
    keys_and_values(AVsForF, As, Vs),!,
    (sort(Vs,[_]) -> fail ; sort(As,ConflSubset)).


mkTriples([],[],[]).
mkTriples([ID|IDs],[List|Lists],Triples) :- 
    (setof(F-(ID-V), member([F,V],List), FAVs);(FAVs=[])),!,
    mkTriples(IDs,Lists,FAVss),
    append(FAVs,FAVss,Triples). 


%provvisorio:
multi_outInitialStateResponse([],_).
multi_outInitialStateResponse([Tck|Tckts],State) :-
    format("  sending response to agent ~w~n",[Tck]),
    m_out(initialStateResponse(Tck,State)),
    multi_outInitialStateResponse(Tckts,State).
%provvisorio:
multi_outIncoherentInitialStateResponse([],_).
multi_outIncoherentInitialStateResponse([Tck|Tckts],Conflicts) :-
    format("  sending response to agent ~w~n",[Tck]),
    m_out(initialStateResponse(Tck,incoherent(Conflicts))),
    multi_outIncoherentInitialStateResponse(Tckts,Conflicts).


%TODO:
%    Al momento il supervisor attende da tutti gli agenti MaxStep azioni da eseguire!!!
%    Sarebbe da prevedere il caso che un agente termini prima (perche' raggiunge il suo goal
%    o perche' abbandona). Per ora assumo che tutti gli agenti che terminano prima di MaxStep passi
%    continui comunque a eseguire NO-OP fino a giungere a MaxStep passi.
%    (altrimenti si blocca in attesa di risposte che non arrivano.... ;) )
performSteps(Tckts,N,MaxSteps,State,State) :- 
    ((Tckts == []) ; (MaxSteps < N)), !,
    m_out(startStep(interactionEnded)).
%format("Supervisor:performSteps ending...  Agents:~w~n", [Tckts]).
performSteps(Tckts,N,MaxSteps,FromState,LastState) :-
    MaxSteps >= N, !,
    %write('STEP ME to next step (type an atom):'),read(_),
    performOneStep(Tckts,N, TerminatedAgents,FromState,ToState),
    format(" Tckts=~w ~n TerminatedAgents=~w ~n",[Tckts,TerminatedAgents]),
%rimuovere/annotare la terminazione di TerminatedAgents
    N1 is N+1,
    findall(Alive, (member(Alive,Tckts), nonmember(Alive,TerminatedAgents)), AliveTckts),
    performSteps(AliveTckts,N1,MaxSteps,ToState,LastState),
    !.

%LAST

    %in progress:   da ActivePriority-Tckt-Conss e ActivePriority-Tckt-(Act-Conss)

performOneStep(Tckts,StepNum,  TerminatedAgents,FromState,ToState) :-
%da completare
    format("Supervisor sending NEXT_STEP atom...~n",[]),
    m_out(startStep(StepNum)),
    format("Supervisor waiting for agents requests...~n",[]),
    getEachAgentsActionsConsequences(Tckts,StepNum,ActionConss,ReorderedTckts),
    format("Supervisor filters active agents conflicts ...~n",[]),
    %write('STEP ME before filterActiveAgents (type an atom):'),read(_),
    filterActiveAgents(ActionConss,ReorderedTckts,ActivePriorityTcktConss,TerminatedAgents),
   format("~n~n~n**************************************************************~n~n~w~n~n~n~n~n~n~n~n",[filterActiveAgents(ActNameMicroDelta_s,ReorderedTckts,ActivePriorityTcktANMDs,TerminatedAgents)]),

    findall(P-T-(Act-Cs), (member(P-T-ListActCons,ActivePriorityTcktConss), member(Act-Cs,ListActCons)), List),
    %amplifyConss(ActivePriorityTcktConss,List),
    sort(List,ActivePTCs),
    %write('~n~n~n******~n~n~nSTEP ME BEFORE conflictReduce (type an atom):'),read(_),nl,
    conflictReduce(ActivePTCs,StepNum,RealActivePTCs,InhibitedPTCs), 
    format("...InhibitedAgents = ~w~n",[InhibitedPTCs]),
    format("...ActiveAgents = ~w~n",[RealActivePTCs]),
    !,
    applyChangesToLocalState(RealActivePTCs,DeltaInState,FromState,ToState),
    m_in(startStep(StepNum)),
    communicateDeltaInState(RealActivePTCs,InhibitedPTCs,StepNum,DeltaInState),
    m_out(completedStep(StepNum)),
    !.

%amplifyConss([],[]).
%amplifyConss([P-T-ListActCons],[]).

conflictReduce([],_,[],[]).
conflictReduce([_P-T-(Act-[])|PriTcktsConss],StepNum,[P-T-(Act-[])|Active],Inhibited) :- !,
    conflictReduce(PriTcktsConss,StepNum,Active,Inhibited).
conflictReduce([P-T-(Act-Cs)|PriTcktsConss],StepNum,Active,Inhibited) :- !,
    conflictingWithThem([P-T-(Act-Cs)], PriTcktsConss, ConflWithSamePri, ConflWithLowerPri, NotConflicting),
    (length(ConflWithSamePri,1) -> (format("Supervisor: conflictcheck on singleton..OK~n",[])) 
    ; (format("Supervisor: waiting for conflict resolution between = ~w~n",[ConflWithSamePri])
    %, write('STEP ME (type an atom):'),read(_),nl
    )),
    conflictResolution_supervisor(ConflWithSamePri, StepNum, Goods, Bads), % *** def in module "SupervisorModuleFile"
    ( (Goods == []) 
        -> ( % unable to obtain a non-conflicting subset
            format("Supervisor: conflict resolution failed.~n",[]),
            conflictReduce(NotConflicting,StepNum,Active,Inhibited1)
           )
        ;  ( %  Goods is the not conflicting subset
            (length(ConflWithSamePri,1) -> true
            ; (format("Supervisor: winning party: ~w~n",[Goods]))),
NotConflicting = SimplifiedNotConflicting,  %OLD,obsoleto...    propagateEffects(Goods, NotConflicting, SimplifiedNotConflicting),
            conflictReduce(SimplifiedNotConflicting,StepNum,Active1,Inhibited1),
            append(Goods, Active1, Active)
           ) ),
           append([ConflWithLowerPri, Bads, Inhibited1], Inhibited),  %si potrebbe ripartire gli Inhibited in "dominated" e "resolved-out" 
    !.


%   propagateEffects([], PriTcktsConss, PriTcktsConss).
%   propagateEffects([_P-_T-(_A-Cs)|Goods], PriTcktsConss, Part) :-
%       removeRedundantPairs(Cs, PriTcktsConss, PriTcktsConss1),
%       propagateEffects(Goods, PriTcktsConss1, Part).
%   
%   removeRedundantPairs([], List, List).
%   removeRedundantPairs([[F,_V]|Cs], PriTcktsConss, Rest) :-
%       removeRedundantPair(F,PriTcktsConss, PriTcktsConss1),
%       removeRedundantPairs(Cs, PriTcktsConss1, Rest).
%   
%   removeRedundantPair(F,PriTcktsConss, PriTcktsConss1) :-
%       findall(P-T-(A-CsNew), (member(P-T-(A-Cs),PriTcktsConss),
%                             findall([F1,V1], (member([F1,V1], Cs),F1\==F), CsNew)),
%           PriTcktsConss1).


%da correggere affinche': se a1 a2,a3 t.c. a1 e a2 conflitto allora a3
%e' in conflitto con a1 anche se a3 concorda con a1 ma confligge con a2
%Questo perche' anche a3 deve partecipare alla coordinazione.
%Altrimenti, se non partecipa, il conflitto persiste se la coordinazione decide per a2

%conflictingWithThem(+ListOfPTCs, +ListOfPTCs1, -ConflWithSamePri, -ConflWithLowerPri, -NotConflicting) 
%ListOfPTCs and ListOfPTCs1 assumed disjoint. Splits ListOfPTCs1 in three lists
%e.g., -ConflWithSamePri is the set of those in ListOfPTCs1 that conflict with some in ListOfPTCs
%or with someone that is in ConflWithSamePri, etc  
conflictingWithThem([], PriTcktsConss, [], [], PriTcktsConss) :-!.
conflictingWithThem([P-T-(Act-Cs)|_], [], [P-T-(Act-Cs)], [], []) :-!.
conflictingWithThem([P-T-(Act-Cs)|Others], PriTcktsConss, ConflWithSamePri, ConflWithLowerPri, NotConflicting) :-
    conflictingWithIt(P,Cs,PriTcktsConss,ConflWithSamePri_wIt,ConflWithLowerPri_wIt,NotConflicting_wIt),
    append(ConflWithSamePri_wIt,Others,MTO),
    sort(MTO,MoreThanOthers),
    conflictingWithThem(MoreThanOthers, NotConflicting_wIt, ConflWithSamePri1, ConflWithLowerPri1, NotConflicting),
    append(ConflWithSamePri_wIt,ConflWithSamePri1,CWSP),
    sort([P-T-(Act-Cs)|CWSP],ConflWithSamePri),
    append(ConflWithLowerPri_wIt,ConflWithLowerPri1,CWLP),
    sort(CWLP,ConflWithLowerPri).

conflictingWithIt(P,Cs, PriTcktsConss, ConflWithSamePri, ConflWithLowerPri, NotConflicting) :-
    tripartite(PriTcktsConss, P, Cs, ConflWithSamePri, ConflWithLowerPri, NotConflicting).

tripartite([], _, _, [], [], []).
tripartite([P1-T1-(Act1-Cs1)|PriNamesTcktsConss], P, Cs, ConflWithSamePri, ConflWithLowerPri, NotConflicting) :-
    theTwoConflict(Cs,Cs1), !,
    %[P1-T1-Cs1|PriNamesTcktsConss] is ordered (w.r.t. P1)
    ((P1 == P) -> ( ConflWithSamePri = [P1-T1-(Act1-Cs1)|CSPs], ConflWithLowerPri = CLPs )
           ;  ( ConflWithSamePri = CSPs, ConflWithLowerPri = [P1-T1-(Act1-Cs1)|CLPs] ) 
    ),
    tripartite(PriNamesTcktsConss, P, Cs, CSPs, CLPs, NotConflicting), !.
tripartite([P1-T1-(Act1-Cs1)|PriTcktsConss], P, Cs, ConflWithSamePri, ConflWithLowerPri, [P1-T1-(Act1-Cs1)|NotConflicting]) :-
    tripartite(PriTcktsConss, P, Cs, ConflWithSamePri, ConflWithLowerPri, NotConflicting).


theTwoConflict(PairsFVs1, PairsFVs2) :-
    member([F,V1], PairsFVs1),
    member([F,V2], PairsFVs2), V1 \== V2.





applyChangesToLocalState(ActivePTCs,DeltaInState,FromState,ToState) :-
    keys_and_values(ActivePTCs,_,ACss),  %P-T-(A-Cs) --> A-Cs
    keys_and_values(ACss,_,Css),  %A-Cs --> Cs
    append(Css, SSS),
    sort(SSS, DeltaInState),
    updateState(DeltaInState, FromState, ToState).

updateState([], State, State):-!.
updateState(State, [], State):-!,
    format("***~nSupervisor: *** WARNING: UNEXPECTED FLUENT NAME(s) ~w  !!!~n***~n",[State]).
updateState([[F,V]|DeltaInState], [[FO,VO]|FromState], [[FO,VO]|ToState]) :-
    FO @< F, !,
    updateState([[F,V]|DeltaInState], FromState, ToState).
updateState([[F,V]|DeltaInState], [[FO,_VO]|FromState], [[F,V]|ToState]) :-
    F == FO, !,
    updateState(DeltaInState, FromState, ToState).
updateState([[F,V]|DeltaInState], [[FO,VO]|FromState], [[F,V]|ToState]) :-
    F @< FO, !,
    format("***~nSupervisor: *** WARNING: UNEXPECTED FLUENT NAME ~w  !!!~n***~n",[[F,V]]),
    updateState(DeltaInState, [[FO,VO]|FromState], ToState).


%DA OTTIMIZZARE: si puo' evitare di spedire DeltaInState per ogni coppia T-A, ma solo per ogni T...
%communicateDeltaInState(+ActivePTCs,+InhibitedPTCs,+StepNum,+DeltaInState)
communicateDeltaInState([],InhibitedPTCs,StepNum,DeltaInState) :-
    communicateDeltaInState(InhibitedPTCs,StepNum,DeltaInState).
communicateDeltaInState([_P-T-(A-_Cs)|ActivePTCs],InhibitedPTCs,StepNum,DeltaInState) :-
    m_out(actionResponse(StepNum,T,success(A,DeltaInState))),
    format("Supervisor: Sending successful response to ~w :~w~n",[T-A,DeltaInState]),
    communicateDeltaInState(ActivePTCs,InhibitedPTCs,StepNum,DeltaInState).

communicateDeltaInState([],_StepNum,_DeltaInState).
communicateDeltaInState([_P-T-(A-_Cs)|InhibitedPTCs],StepNum,DeltaInState) :-
    m_out(actionResponse(StepNum,T,conflicting(A,DeltaInState))),
    format("Supervisor: Sending response (inhibition) to ~w (delta=~w)~n",[T-A,DeltaInState]),
    communicateDeltaInState(InhibitedPTCs,StepNum,DeltaInState).


%TODO controllare filterActiveAgents: versione preliminare (ridondante/lacunosa...)
    %act([single],act1)-[[f1,10]]
    % POSSIBILI AZIONI : 
    % 'NO-OP'    =  l'agente in questo turno attende
    % 'NO-MORE'  =  l'agente ha terminato
    filterActiveAgents([],_Tckts,[],[]).
    filterActiveAgents(['NO-OP'|ActionActConss],[_Tck|Tckts],PriTcktsConss,TerminatedAgents) :-!,
        filterActiveAgents(ActionActConss,Tckts,PriTcktsConss,TerminatedAgents).
    filterActiveAgents([act([single],'NO-OP')|ActionActConss],[_Tck|Tckts],PriTcktsConss,TerminatedAgents) :-!,
        filterActiveAgents(ActionActConss,Tckts,PriTcktsConss,TerminatedAgents).
    filterActiveAgents([act([single],'NO-OP')-_|ActionActConss],[_Tck|Tckts],PriTcktsConss,TerminatedAgents) :-!,
        filterActiveAgents(ActionActConss,Tckts,PriTcktsConss,TerminatedAgents).
    filterActiveAgents(['NO-MORE'|ActionActConss],[Tck|Tckts],PriTcktsConss,[Tck|TerminatedAgents]) :-!,
        filterActiveAgents(ActionActConss,Tckts,PriTcktsConss,TerminatedAgents).
    filterActiveAgents([act([single],'NO-MORE')|ActionActConss],[Tck|Tckts],PriTcktsConss,[Tck|TerminatedAgents]) :-!,
        filterActiveAgents(ActionActConss,Tckts,PriTcktsConss,TerminatedAgents).
    filterActiveAgents([act([single],'NO-MORE')-_|ActionActConss],[Tck|Tckts],PriTcktsConss,[Tck|TerminatedAgents]) :-!,
        filterActiveAgents(ActionActConss,Tckts,PriTcktsConss,TerminatedAgents).
    filterActiveAgents([ActCons|ActionActConss],[Tck|Tckts],[Pri-Tck-(ActCons)|PriTcktsConss],TerminatedAgents) :-!,
    %ActCons e' Act-Cons con Cons lista di coppie [Flu,Val]
        getTheAgentPriority(Tck,Pri),
        filterActiveAgents(ActionActConss,Tckts,PriTcktsConss,TerminatedAgents).

getEachAgentsActionsConsequences(Tckts,StepNum,MActions,ReorderedTckts) :-
    %MActions e' lista di elementi che sono NO-MORE oppure lista di ActName-MicroDelta per 
    %il corrisp. Tck in ReorderedTckts
    %read(_),
    %write('STEP ME (type an atom):'),read(_),
    expectedTermsTer(Tckts, List, StepNum, proposedAction(StepNum,_Tck,_Term)),
    getMicroActions(Tckts, List, MActions, StepNum,ReorderedTckts).
%expectedTermsQua_list(ReorderedTckts, MActionCounts, Lists, StepNum, proposedAction(StepNum,_Tck,_N,_AD)),
%   getActionConss(ReorderedTckts, Lists, ActionMiDeltas, StepNum, ReReorderedTckts).


getMicroActions([], _List, [], _, []).
getMicroActions(Tckts, List, [PropA|PropAs], StepNum, [ATck|RTckts]) :-
    %format("~n*~n*~n*~n*~n*~n*~n*~n*~n getMicroActions call : ~w~n", [getMicroActions(Tckts, List, [PropA|PropAs], StepNum, [ATck|RTckts])]),
    %m_in(List, proposedAction(StepNum,ATck,APairs)),
    %format("        RECEIVE ACTION = ~w~n", [proposedAction(StepNum,ATck,APairs)]),
    m_in(List, proposedAction(StepNum,ATck,PropA)),
    %format("        RECEIVE ACTION = ~w~n", [proposedAction(StepNum,ATck,PropA)]),
    selectchk(ATck, Tckts, TcktsRest),  % PropA e' NO-MORE oppure lista di ActName-MicroDelta per ATck
    getMicroActions(TcktsRest, List, PropAs, StepNum, RTckts).


%getActionConss([], _Lists, [], _, []).
%getActionConss([ATck|Tckts], [List|Lists], AConsList, StepNum, ATcklist) :-
%        getActionConss_aux(ATck, List, ATcklist1, AConsList1, StepNum),
%   getActionConss(Tckts, Lists, AConsList2, StepNum, ATcklist2),
%   append(AConsList1, AConsList2, AConsList),
%   append(ATcklist1, ATcklist2, ATcklist).
%
%%act([single],act2)-[[f2,22]]
%getActionConss_aux(_ATck, [], [], [], _StepNum).
%getActionConss_aux(ATck, [Term|Terms], [ATck|ATcklist], [(ActName-MicroDelta)|ConsList], StepNum) :-
%   m_in([Term], proposedAction(StepNum,AgentTck,N,ActName-MicroDelta)),
%   format("        RECEIVE ACTION = ~w~n", [proposedAction(StepNum,AgentTck,N,ActName-MicroDelta)]),
%   getActionConss_aux(ATck, Terms, ATcklist, ConsList, StepNum).


%OLD
%getActionConss([], _Lists, [], _, []).
%getActionConss(Tckts, List, [ActionNameMulti|ActionNameMultiList], StepNum, [ATck|RTckts]) :-
%   %m_in(List, proposedAction(StepNum,ATck,APairs)),
%   %format("        RECEIVE ACTION = ~w~n", [proposedAction(StepNum,ATck,APairs)]),
%   %m_in(List, proposedAction(StepNum,ATck,ActionNameMulti)),
%   m_in(List, proposedAction(StepNum,AgentTck,N,ActName-MicroDelta)),
%   format("        RECEIVE ACTION = ~w~n", [proposedAction(StepNum,AgentTck,N,ActName-MicroDelta)]),
%   selectchk(ATck, Tckts, TcktsRest),
%   getActionConss(TcktsRest, List, ActionNameMultiList, StepNum, RTckts).


expectedTermsTer([], [], _StepNum, _Template).
expectedTermsTer([Tck|Tckts],[Term|R],StepNum, Template) :-
    copy_term(Template,Term),
    Term =.. [_,StepNum,Tck|_],
    expectedTermsTer(Tckts, R, StepNum, Template).




endingStuff(Tckts,FinalState) :-
%da fare
format("Supervisor:endingStuff Agents:~w~n", [Tckts]),
format("Supervisor:endingStuff FinalState:~w~n", [FinalState]).


closeSession :-
    m_shutdown_server,
    m_close_client.%,
    %   halt.




